import os

import cv2
import numpy as np
from loguru import logger
from numpy.typing import NDArray
from typing import Tuple, List, Optional

from ..Consts.Types import *
from .Progress import progress_bar


# video utils 
class VideoSaver():
    def __init__(self, name: str, size: Union[Tuple[int,int], Literal['auto']] = 'auto', frame_rate: float = 15.0):
        self.name = name
        self.frame_rate = frame_rate
        self.fourcc = cv2.VideoWriter_fourcc(*'XVID') # type: ignore
        self.size = size 

        if size=='auto':
            self.vidwriter = None 
        else:
            self.vidwriter = cv2.VideoWriter(self.name, self.fourcc, self.frame_rate, (self.size[1],self.size[0]))

    def write(self, img: NDArray):
        if self.vidwriter is None:
            self.size = [img.shape[0], img.shape[1]]
            self.vidwriter = cv2.VideoWriter(self.name, self.fourcc, self.frame_rate,(self.size[1],self.size[0]))
        assert (img.shape[0]==self.size[0]) and (img.shape[1]==self.size[1]), f'VideoSaver: Image shape is not consistent. Set {self.size}, but got {img.shape}'
        self.vidwriter.write(img)

    def finish(self):
        self.vidwriter.release()


def check_frame_num(fname: str) -> int:
    video = cv2.VideoCapture(fname)
    framenum = int(video.get(cv2.CAP_PROP_FRAME_COUNT))
    return framenum


def check_fps(fname: str) -> float:
    video = cv2.VideoCapture(fname)
    fps = video.get(cv2.CAP_PROP_FPS)
    return fps 


def combine_audio(vidname: str, audname: str, outname: str, fps: Optional[float]=None) -> None:
    import moviepy.editor as mpe
    my_clip = mpe.VideoFileClip(vidname)
    audio_background = mpe.AudioFileClip(audname)
    final_clip = my_clip.set_audio(audio_background)
    final_clip.write_videofile(outname,fps=fps)


def compress_video(vidname: str, outname: str, fps: Optional[float]=None) -> None:
    import moviepy.editor as mpe
    my_clip = mpe.VideoFileClip(vidname)
    my_clip.write_videofile(outname, fps=fps)


def extract_frames(fname: str, output_dir: str, ext: BasicImageTypes='jpg', skip: int=0, frame_format: str='frame_%08d', return_images: bool=False) -> List[NDArray[np.uint8]]:
    assert isinstance(output_dir, str), 'output_dir must be string'
    assert isinstance(fname, str), 'file name must be string'
    assert ext.lower() in ['jpg', 'jpeg', 'png'], "extension must be one of ['jpg', 'jpeg', 'png']"

    logger.info(f'Extracting {fname}\tOutput_dir:{output_dir}\tFormat:{frame_format}\tExtension:{ext}\tSkip:{skip}')
    os.makedirs(output_dir, exist_ok=True)
    cap = cv2.VideoCapture(fname)
    framenum = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))

    cnt = 0
    images = []
    with progress_bar() as prog:
        task = prog.add_task(description='Extracting...', total=framenum)
        while 1:
            ret, frame = cap.read()
            if not ret:
                break
            if cnt%(skip+1)==0:
                if return_images:
                    images.append(frame)
                else:
                    imgname = os.path.join(output_dir, frame_format%cnt + '.' + ext)
                    cv2.imwrite(imgname, frame)
            cnt += 1 
            prog.advance(task)

    logger.info('Extraction finished.')
    return images



import torch 
import torch.nn.functional as F 
from ..Base import Model 
from typing import List 
from .Quantizers import QuantizerBase
from .Observers import PlaceholderObserver
from torch import Tensor 
from collections.abc import Callable
import itertools 
from loguru import logger 


def l2_score(a, b):
    return -torch.pow(a - b, 2).mean()


def cosine_score(a, b):
    a = a.flatten(1)
    b = b.flatten(1)
    dist = F.cosine_similarity(a, b, dim=1)
    return dist.mean()


class LayerCalibrator(Model):
    def initialize(self, quantizers: List[QuantizerBase], fwd_fn: Callable[..., Tensor]):
        self.observers = []
        for q in quantizers:
            self.observers.append(q.observer)
        for obs in self.observers:
            assert isinstance(obs, PlaceholderObserver), 'Observer for LayerCalibrator must be PlaceholderObserver'
        self.inputs = []
        self.outputs = []
        self.fwd_fn = fwd_fn

    def layer_hook(self, module, inputs, outputs):
        if self._quant and self._quant_calibrating:
            self.inputs.append(inputs)
            self.outputs.append(outputs)

    @torch.no_grad()
    def _finish_calibrate(self):
        if len(self.inputs)==0:
            return 
        scr_func_flag = self.get_flag('CalibScoreFunc')
        sim_func = cosine_score if scr_func_flag=='cosine' else l2_score
        logger.debug(f'LayerCalibrator post processing... Score function: {"L2" if scr_func_flag is None else scr_func_flag}')
        # combine inputs 
        logger.debug(f'Num inputs to calibrate: {len(self.inputs)}')
        inputs = list(zip(*self.inputs))
        inputs = [torch.cat(i, dim=0) for i in inputs]
        outputs = torch.cat(self.outputs, dim=0)

        # initialize observers
        for obs in self.observers:
            obs.init_quant_params()

        # phase 1: generate coarse grid to search best scales  
        scales_to_test = [[s*0.05+0.2 for s in range(17)],] * len(self.observers)

        best_scales = None 
        best_sim = None 

        for scales in itertools.product(*scales_to_test):
            for s,obs in zip(scales, self.observers):
                obs.resize_scale(s)
            quant_output = self.fwd_fn(*inputs)
            similarity = sim_func(quant_output, outputs)
            if best_sim is None or similarity>best_sim:
                best_scales = scales
                best_sim = similarity

        # phase 2: generate fine grid to search best scales 
        scales_to_test: List[List[float]] = [[s-0.05+0.002*i for i in range(51)] for s in best_scales]  # type: ignore
        
        best_scales = None 
        best_sim = None
        for scales in itertools.product(*scales_to_test):
            for s,obs in zip(scales, self.observers):
                obs.resize_scale(s)
            quant_output = self.fwd_fn(*inputs)
            similarity = sim_func(quant_output, outputs)
            if best_sim is None or similarity>best_sim:
                best_scales = scales
                best_sim = similarity

        for s,obs in zip(best_scales, self.observers):  # type: ignore
            obs.resize_scale(s)
        logger.debug(f'Find best scale: {best_scales}. Similarity: {best_sim.cpu().numpy()}')  # type: ignore

    def _load_from_state_dict(self, state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys, error_msgs):
        ...

    def _save_to_state_dict(self, destination, prefix, keep_vars):
        ...




import math
from typing import Union, List, Tuple

import torch
import torch.nn.init as init
import torchvision.ops as ops
from torch import Tensor
from torch.nn.parameter import Parameter

from ..Base import Model
from ..Consts.Types import *


## Resnet style initialization 
def _resnet_normal(tensor):
    fan_in, fan_out = init._calculate_fan_in_and_fan_out(tensor)
    std = math.sqrt(2.0 / float(fan_out))
    return init._no_grad_normal_(tensor, 0., std)


class DeformConv2D(Model):
    TypeKSize = Union[int, List[int], Tuple[int,int]]
    size: TypeKSize
    kernel_size = Tuple[int,int,int,int]
    outchn: int 
    stride: TypeKSize 
    pad_mode: PadModes
    dilation_rate: int 
    usebias: bool
    pad: Union[int, Tuple[int,int]]
    groups: int
    inchn: int
    weight: Parameter
    bias: Union[Parameter, None]
    
    def __init__(self, size:TypeKSize, outchn:int, stride:TypeKSize=1, pad:PadModes='SAME_LEFT', dilation_rate:int=1, usebias:bool=True):
        super().__init__(size=size, outchn=outchn, stride=stride, pad=pad, dilation_rate=dilation_rate, usebias=usebias)

    def initialize(self, size:TypeKSize, outchn:int, stride:TypeKSize=1, pad:PadModes='SAME_LEFT', dilation_rate:int=1, usebias:bool=True):
        self.size = size
        self.outchn = outchn
        self.stride = stride
        self.usebias = usebias
        self.dilation_rate = dilation_rate
        assert (pad in ['VALID','SAME_LEFT'])
        self.pad_mode = pad 

    def _parse_args(self, input_shape: List[int]):
        inchannel = input_shape[1]
        self.inchn = inchannel
        # parse args
        if isinstance(self.size, list) or isinstance(self.size, tuple):
            if self.pad_mode == 'VALID':
                self.pad = 0
            else:
                self.pad = ((self.size[0]+ (self.dilation_rate-1) * ( self.size[0]-1 ))//2, (self.size[1]+ (self.dilation_rate-1) * (self.size[1]-1))//2)
            self.kernel_size = (self.outchn, inchannel // self.groups, self.size[0], self.size[1])
        elif isinstance(self.size, int):
            if self.pad_mode == 'VALID':
                self.pad = 0
            else:
                self.pad = (self.size + (self.dilation_rate-1) * (self.size-1))//2
            self.kernel_size = (self.outchn, inchannel // self.groups, self.size, self.size)

    def build(self, *inputs):
        inp = inputs[0]
        self._parse_args(inp.shape)
        self.weight = Parameter(torch.Tensor(*self.size))
        if self.usebias:
            self.bias = Parameter(torch.Tensor(self.outchn))
        else:
            self.register_parameter('bias', None)
        self.reset_params()

    def reset_params(self):
        if self.get_flag('conv_init_mode')=='normal':
            init.normal_(self.weight, std=0.001)
        elif self.get_flag('conv_init_mode')=='resnet':
            _resnet_normal(self.weight)
        else:
            init.kaiming_uniform_(self.weight, a=math.sqrt(5))  # type: ignore
        if self.bias is not None:
            if self.get_flag('conv_init_mode')=='normal':
                init.zeros_(self.bias)
            elif self.get_flag('conv_init_mode')=='resnet':
                fan_in, _ = init._calculate_fan_in_and_fan_out(self.weight)
                bound = 1 / math.sqrt(fan_in)
                init.uniform_(self.bias, -bound, bound)
            else:
                fan_in, _ = init._calculate_fan_in_and_fan_out(self.weight)
                bound = 1 / math.sqrt(fan_in)
                init.uniform_(self.bias, -bound, bound)

    def forward(self, x: Tensor, offset: Tensor, mask: Union[Tensor,None]=None) -> Tensor:
        return ops.deform_conv2d(x, offset, self.weight, self.bias, self.stride, self.pad, self.dilation_rate, mask)       # type: ignore

    def __call__(self, x: Tensor, offset: Tensor, mask: Union[Tensor,None]=None) -> Tensor:
        return super().__call__(x, offset, mask)




import copy
import inspect
from typing import Any, Dict

import torch
import torch.nn as nn
from loguru import logger

from ..Modules import ConvLayer, ConvLayer1D, ConvLayer3D, DeConvLayer, Dense
from ..Quant import QAct


def init_model(model, *args, **kwargs):
    # run one forward loop to initialize model 
    with torch.no_grad():
        return model(*args, **kwargs)


def to_standard_torch(model: nn.Module, inplace: bool=True) -> nn.Module:
    if not inplace:
        model = copy.deepcopy(model)
    if hasattr(model, 'to_torch_module'):
        model.to_torch_module()  # type: ignore
    else:
        for n,c in model.named_children():
            if hasattr(c, 'to_torch'):
                setattr(model, n.split('.')[-1], c.to_torch())  # type: ignore
            else:
                to_standard_torch(c)
    return model 


def inspect_quant_params(module: nn.Module, result_dict: Dict[str, Any]=dict(), prefix: str=''):
    if isinstance(module, QAct):
        try:
            zero_point = module.quantizer.observer.zero_point
            scale = module.quantizer.observer.scale 
            result_dict[prefix] = [scale, zero_point.round()] # type: ignore
        except:
            logger.warning(f'Quant params of layer: {prefix} cannot be properly retrieved. Maybe this layer is never called in calibration.')
        return result_dict
    if isinstance(module, (ConvLayer, ConvLayer1D, ConvLayer3D)):
        try:
            if hasattr(module.conv, 'input_quantizer'):
                scale = module.conv.input_quantizer.observer.scale
                zero_point = module.conv.input_quantizer.observer.zero_point
                result_dict[prefix+'/conv/Conv__input'] = [scale, zero_point.round()] # type: ignore
                scale = module.conv.w_quantizer.observer.scale
                zero_point = module.conv.w_quantizer.observer.zero_point
                result_dict[prefix+'/conv/Conv__weight'] = [scale, zero_point.round()] # type: ignore
        except:
            logger.warning(f'Quant params of layer: {prefix} cannot be properly retrieved. Maybe this layer is never called in calibration.')
        return result_dict
    if isinstance(module, DeConvLayer):
        try:
            if hasattr(module.conv, 'input_quantizer'):
                scale = module.conv.input_quantizer.observer.scale
                zero_point = module.conv.input_quantizer.observer.zero_point
                result_dict[prefix+'/conv/DeConv__input'] = [scale, zero_point.round()] # type: ignore
                scale = module.conv.w_quantizer.observer.scale
                zero_point = module.conv.w_quantizer.observer.zero_point
                result_dict[prefix+'/conv/DeConv__weight'] = [scale, zero_point.round()] # type: ignore
        except:
            logger.warning(f'Quant params of layer: {prefix} cannot be properly retrieved. Maybe this layer is never called in calibration.')
        return result_dict
    if isinstance(module, Dense):
        try:
            if hasattr(module.fc, 'input_quantizer'):
                scale = module.fc.input_quantizer.observer.scale
                zero_point = module.fc.input_quantizer.observer.zero_point
                result_dict[prefix+'/fc/Dense__input'] = [scale, zero_point.round()] # type: ignore
                scale = module.fc.w_quantizer.observer.scale
                zero_point = module.fc.w_quantizer.observer.zero_point
                result_dict[prefix+'/fc/Dense__weight'] = [scale, zero_point.round()] # type: ignore
        except:
            logger.warning(f'Quant params of layer: {prefix} cannot be properly retrieved. Maybe this layer is never called in calibration.')
        return result_dict
    if isinstance(module, (QAdd, QMatmul)):
        try:
            if hasattr(module, 'a_quantizer'):
                scale = module.a_quantizer.observer.scale
                zero_point = module.a_quantizer.observer.zero_point
                result_dict[prefix+'/input__1'] = [scale, zero_point.round()] # type: ignore
                scale = module.b_quantizer.observer.scale
                zero_point = module.b_quantizer.observer.zero_point
                result_dict[prefix+'/input__2'] = [scale, zero_point.round()] # type: ignore
        except:
            logger.warning(f'Quant params of layer: {prefix} cannot be properly retrieved. Maybe this layer is never called in calibration.')
        return result_dict
    if isinstance(module, (nn.ModuleList, nn.Sequential)):
        for i in range(len(module)):
            inspect_quant_params(module[i], result_dict=result_dict, prefix=prefix+'.%d'%i)
        return result_dict
    results = inspect.getmembers(module)
    for name, child_module in results:
        if isinstance(child_module, nn.Module):
            inspect_quant_params(child_module, result_dict=result_dict, prefix=prefix+'/'+name)
    return result_dict




from .Activation import Activation
from .Conv import conv1D, conv2D, conv3D, deconv2D
from .Extension import DeformConv2D
from .FC import fcLayer
from .Norm import BatchNorm, LayerNorm
from .Pool import AvgPool2d, MaxPool2d
from .Upsample import BilinearUpSample, NNUpSample



from abc import ABC, abstractmethod
from typing import Type, Union, Tuple, Dict, List

import torch
from loguru import logger
from torch import Tensor
from torch.nn.parameter import Parameter

from ..Base import Model
from ..Consts.Types import *
from .QTypes import QTypeBase


##### START: Observer classes 
class ObserverBase(Model, ABC):
    min_val: Union[Tensor, None]
    max_val: Union[Tensor, None]
    mode: QuantModes
    bit_type: Type[QTypeBase]
    zero_offset: bool
    is_weight: bool
    scale: Union[Tensor, None]
    zero_point: Union[Tensor, None]
    dim: int

    def __init__(self, bit_type: Type[QTypeBase], zero_offset: bool, mode: QuantModes, is_weight: bool):
        super().__init__(bit_type, zero_offset, mode, is_weight)
        assert mode in ['layer_wise', 'channel_wise']
        self.min_val = None 
        self.max_val = None 
        self.mode = mode 
        self.bit_type = bit_type
        self.zero_offset = zero_offset
        self.is_weight = is_weight
        self.scale = None 
        self.zero_point = None 

    def build(self, x: Tensor):
        if self.is_weight:
            self.dim = 0
        else:
            if len(x.shape)==4:
                self.dim = 1 
            else:
                self.dim = -1

    @abstractmethod
    def observe(self, x: Tensor):
        ...

    @torch.no_grad()
    def get_quant_params(self, max_val: Tensor, min_val: Tensor) -> Tuple[Tensor, Tensor]:
        if self.zero_offset:
            max_val = torch.max(max_val, -min_val)
            scale = max_val / min(self.bit_type.max_val, -self.bit_type.min_val)
            zero_point = torch.zeros_like(max_val)
        else:
            scale = (max_val - min_val) / float(self.bit_type.max_val - self.bit_type.min_val)
            zero_point = self.bit_type.min_val - torch.round(self.min_val / scale)
        scale.clamp(torch.finfo(torch.float32).eps)
        return scale, zero_point

    def _finish_calibrate(self):
        if self.scale is None:
            if (self.min_val is None) or (self.max_val is None):
                logger.warning('This quant layer is not fed with any data and it will be omitted. Use M.inspect_quant_params to get specific layer name')
            elif self.scale is None:
                s,z = self.get_quant_params(self.max_val, self.min_val)
                self.scale = Parameter(s)
                self.zero_point = Parameter(z)

    def forward(self, x: Tensor) -> Tensor:
        if self._quant_calibrating:
            self.observe(x)
        return x 
    
    def __call__(self, x: Tensor) -> Tensor:
        return super().__call__(x)

    def _load_from_state_dict(self, state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys, error_msgs):
        if prefix+'scale' in state_dict:
            self.scale = Parameter(state_dict[prefix + 'scale'])
            self.zero_point = Parameter(state_dict[prefix + 'zero_point'])
        else:
            logger.trace(f'No scale in checkpoint for layer [{prefix}]')

    def _save_to_state_dict(self, destination, prefix, keep_vars):
        if self._quant_calibrated:
            destination[prefix + 'scale'] = self.scale 
            destination[prefix + 'zero_point'] = self.zero_point


class PlaceholderObserver(ObserverBase):
    def observe(self, x):
        if self.mode == 'channel_wise':
            x = x.transpose(0, self.dim)
            x = x.flatten(1)
            minv = x.min(dim=1)[0]
            maxv = x.max(dim=1)[0]
        else:
            minv = x.min()
            maxv = x.max()
        if (self.min_val is None) or (self.max_val is None):
            self.min_val = minv
            self.max_val = maxv
        else:
            self.min_val = torch.minimum(self.min_val, minv)
            self.max_val = torch.maximum(self.max_val, maxv)
            
    def initialize(self, *args, **kwargs):
        self.init_scale = None 
        self.zero_point = None 

    def init_quant_params(self):
        self.init_scale, self.zero_point = self.get_quant_params(self.max_val, self.min_val)

    def resize_scale(self, factor):
        self.scale = Parameter(factor * self.init_scale)
        self.zero_point = Parameter(self.zero_point)
        
    def _finish_calibrate(self):
        ...


class PercentileObserver(ObserverBase):
    sigma: float
    percentile: float

    def initialize(self, *args, **kwargs):
        self.sigma = 0.01 
        self.percentile = 0.99999

    def observe(self, x):
        if self.mode == 'channel_wise':
            x = x.transpose(0, self.dim)
            x = x.flatten(1)
            minv = torch.quantile(x.cuda(), 1-self.percentile, dim=1).cpu()
            maxv = torch.quantile(x.cuda(), self.percentile, dim=1).cpu()
        else:
            minv = torch.quantile(x.cuda(), 1-self.percentile).cpu()
            maxv = torch.quantile(x.cuda(), self.percentile).cpu()
        if (self.min_val is None) or (self.max_val is None):
            self.min_val = minv
            self.max_val = maxv
        else:
            self.min_val = self.sigma * minv + (1 - self.sigma) * self.min_val
            self.max_val = self.sigma * maxv + (1 - self.sigma) * self.max_val


class MinMaxObserver(ObserverBase):
    def observe(self, x):
        if self.mode == 'channel_wise':
            x = x.transpose(0, self.dim)
            x = x.flatten(1)
            minv = x.min(dim=1)[0]
            maxv = x.max(dim=1)[0]
        else:
            minv = x.min()
            maxv = x.max()
        if (self.min_val is None) or (self.max_val is None):
            self.min_val = minv
            self.max_val = maxv
        else:
            self.min_val = torch.minimum(self.min_val, minv)
            self.max_val = torch.maximum(self.max_val, maxv)


class OmseObserver(ObserverBase):
    x_buffer: List[Tensor]

    def initialize(self, *args, **kwargs):
        self.x_buffer = []

    def observe(self, x):
        if self.mode == 'channel_wise':
            x = x.transpose(-1, self.dim)
            x = x.flatten(0,-2)
            minv = x.min(dim=0)[0]
            maxv = x.max(dim=0)[0]
        else:
            minv = x.min()
            maxv = x.max()
        if (self.min_val is None) or (self.max_val is None):
            self.min_val = minv
            self.max_val = maxv
        else:
            self.min_val = torch.minimum(self.min_val, minv)
            self.max_val = torch.maximum(self.max_val, maxv)

        self.x_buffer.append(x)

    @torch.no_grad()
    def _least_square(self, max_val: Tensor, min_val: Tensor, scale: Tensor, factor: float):
        if self.zero_offset:
            zero_point = torch.zeros_like(max_val)
        else:
            zero_point = self.bit_type.min_val - torch.round(min_val * factor / scale)

        scale = scale.cuda()
        zero_point = zero_point.cuda()
        x_buffer = torch.cat(self.x_buffer, dim=0).cuda()

        scale = scale * factor
        x_buffer_q = x_buffer / scale + zero_point
        x_buffer_q = x_buffer_q.round().clamp(self.bit_type.min_val, self.bit_type.max_val)
        x_buffer_q = (x_buffer_q - zero_point) * scale 
        return torch.pow(x_buffer_q - x_buffer, 2).mean(), scale, zero_point

    @torch.no_grad()
    def get_quant_params(self, max_val: Tensor, min_val: Tensor):
        if self.zero_offset:
            max_val = torch.max(max_val, -min_val)
            scale = max_val / min(self.bit_type.max_val, -self.bit_type.min_val)
        else:
            scale = (max_val - min_val) / float(self.bit_type.max_val - self.bit_type.min_val)
        scale.clamp(torch.finfo(torch.float32).eps)
        min_score, scale_best, zero_point_best = self._least_square(max_val, min_val, scale, 1.0)
        for i in range(50):
            factor = 1 - 0.01*i
            score, scale_i, zero_point_i = self._least_square(max_val, min_val, scale, factor)
            if min_score is None:
                min_score = score
                scale_best = scale_i 
                zero_point_best = zero_point_i
            if score<min_score:
                min_score = score
                scale_best = scale_i
                zero_point_best = zero_point_i
        return scale_best, zero_point_best


QObservers: Dict[Union[QObserverTypes, str], Type[ObserverBase]] = \
        {"minmax": MinMaxObserver, 'percentile': PercentileObserver, 'omse': OmseObserver, 'placeholder': PlaceholderObserver}
##### END: Observer classes 




import functools
import os
from typing import Any, Union

import torch
import torch.nn as nn
from torch import Tensor

from .Consts.Types import PRESET_FLAGS


class Model(nn.Module):
    def __init__(self, *args, **kwargs) -> None:
        super(Model, self).__init__()
        self._quant_calibrating = False
        self._quant_calibrated = False
        self._quant = False
        self._is_built = False
        self._model_flags = {}
        self.initialize(*args, **kwargs)

    def initialize(self, *args, **kwargs) -> None:
        pass 

    def _set_status(self) -> None:
        if self._quant:
            self.start_quant()
        for k in self._model_flags:
            self.set_flag(k, self._model_flags[k])
            
    def build(self, *inputs, **kwargs) -> None:
        pass

    def init_params(self, *inputs, **kwargs):
        pass 

    def __call__(self, *input, **kwargs):
        if not self._is_built:
            self.build(*input)
            self._set_status()
        for hook in self._forward_pre_hooks.values():
            result = hook(self, input)
            if result is not None:
                if not isinstance(result, tuple):
                    result = (result,)
                input = result
        if torch._C._get_tracing_state():                            # type: ignore
            result = self._slow_forward(*input, **kwargs)
        else:
            result = self.forward(*input, **kwargs)
            if not self._is_built:
                self.init_params(*input, **kwargs)
        for hook in self._forward_hooks.values():
            hook_result = hook(self, input, result)
            if hook_result is not None:
                result = hook_result
        if len(self._backward_hooks) > 0:
            var = result
            while not isinstance(var, torch.Tensor):
                if isinstance(var, dict):
                    var = next((v for v in var.values() if isinstance(v, torch.Tensor)))
                else:
                    var = var[0]
            grad_fn = var.grad_fn
            if grad_fn is not None:
                for hook in self._backward_hooks.values():
                    wrapper = functools.partial(hook, self)
                    functools.update_wrapper(wrapper, hook)
                    grad_fn.register_hook(wrapper)
        self._is_built = True
        return result

    def set_flag(self, k: Union[PRESET_FLAGS,Any], v: Any=True) -> None:
        def set_model_flag(obj):
            if hasattr(obj, '_model_flags'):
                obj._model_flags[k] = v 
        self.apply(set_model_flag)

    def get_flag(self, k):
        return self._model_flags.get(k, None)

    def start_calibrate(self):
        def set_calibarte(obj):
            if hasattr(obj, '_quant') and obj._quant:
                obj._quant_calibrating = True
        self.apply(set_calibarte)

    def end_calibrate(self):
        def unset_calibrate(obj):
            if hasattr(obj, '_quant') and obj._quant:
                obj._quant_calibrating = False
                if hasattr(obj, '_finish_calibrate'):
                    obj._finish_calibrate()
                obj._quant_calibrated = True
        self.apply(unset_calibrate)

    def start_quant(self):
        def set_quant(obj):
            obj._quant = True
        self.apply(set_quant)

    def end_quant(self):
        def unset_quant(obj):
            obj._quant = False
        self.apply(unset_quant)

    def save_tensor(self, out: Tensor, name: str):
        if self.get_flag('save_tensor'):
            os.makedirs('./layer_dumps/', exist_ok=True)
            torch.save(out, './layer_dumps/%s.pth'%name.replace('/','_'))

    def _load_from_state_dict(self, state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys, error_msgs):
        try:
            self._load_from_state_dict2(state_dict, prefix)
        except Exception as e:
            if not self.get_flag('loose_load'):
                raise e 
            else:
                pass
        super()._load_from_state_dict(state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys, error_msgs)

    def _load_from_state_dict2(self, state_dict, prefix):
        # Conveinient method. Omit infrequent arguments
        pass 



import torch.nn as nn
from loguru import logger
from torch import Tensor

from .. import Layers as L
from ..Base import Model
from ..Consts.Activate import *
from ..Consts.Types import *
from ..Quant import QuantizerBase


class ConvLayer(Model):
    conv: L.conv2D
    batch_norm: bool 
    bn: L.BatchNorm
    act: L.Activation

    def __init__(self, size: TypeKSize2D, outchn:int, stride:TypeKSize2D=1, pad:PadModes='SAME_LEFT', dilation_rate:int=1, activation:int=-1,
                 batch_norm:bool=False, affine:bool=True, usebias:bool=True, groups:int=1):
        super().__init__(size=size, outchn=outchn, stride=stride, pad=pad, dilation_rate=dilation_rate, activation=activation,
                         batch_norm=batch_norm, affine=affine, usebias=usebias, groups=groups)
        
    def initialize(self, size: TypeKSize2D, outchn:int, stride:TypeKSize2D=1, pad:PadModes='SAME_LEFT', dilation_rate:int=1, activation:int=-1,
                 batch_norm:bool=False, affine:bool=True, usebias:bool=True, groups:int=1):
        self.conv = L.conv2D(size, outchn, stride, pad, dilation_rate, usebias, groups)
        if batch_norm:
            self.bn = L.BatchNorm(affine=affine)
        self.batch_norm = batch_norm
        self.act = L.Activation(activation)

    def forward(self, x) -> Tensor:
        x = self.conv(x)
        if self.batch_norm:
            x = self.bn(x)
        x = self.act(x)
        return x 
    
    def __call__(self, x: Tensor) -> Tensor:
        return super().__call__(x)

    def to_torch_module(self):
        conv = self.conv.to_torch()
        setattr(self, 'conv', conv)
        if self.batch_norm:
            bn = self.bn.to_torch()
            setattr(self, 'bn', bn)
        if self.activation==PARAM_RELU:
            relu = nn.ReLU()
            setattr(self, 'act', relu)

    def get_input_quantizer(self) -> QuantizerBase:
        return self.conv.input_quantizer

    def get_weight_quantizer(self) -> QuantizerBase:
        return self.conv.w_quantizer

    def set_input_quantizer(self, quantizer: QuantizerBase):
        self.conv.set_input_quantizer(quantizer)

    def _load_from_state_dict2(self, state_dict, prefix):
        def _load_weight(k):
            if not k in state_dict:
                raise KeyError(f'Attenpt to find [{k}] but only exist [{state_dict.keys()}] Cannot find weight in checkpoint for layer: [{prefix}]')
            return state_dict.pop(k)
        def _load_bias(k):
            if self.conv.usebias:
                try:
                    b = state_dict.pop(k)
                except:
                    raise KeyError(f'Attenpt to find [{k}] but only exist [{state_dict.keys()}] Bias is set for layer [{prefix}] but not found in checkpoint.')
            else:
                b = None
            return b 
        
        if not self._is_built:
            logger.warning(f'Layer: [{prefix}] is not built. This layer was not used in dummy forward. Skipping loading state_dict.')
            return 

        # get names for params
        if prefix+'conv.weight' in state_dict:
            # normal load 
            w = prefix + 'conv.weight'
            b = prefix + 'conv.bias'
        elif self.get_flag('fc2conv') and ((prefix+'fc.weight') in state_dict):
            w = prefix + 'fc.weight'
            b = prefix + 'fc.bias'
        elif self.get_flag('from_torch'):
            w = prefix + 'weight'
            b = prefix + 'bias'
        else:
            raise KeyError(f'Cannot find weight in checkpoint for layer: [{prefix}]')

        # laod weight and bias 
        w = _load_weight(w)
        b = _load_bias(b)
        if self.get_flag('fc2conv') and len(w.shape)==2:
            w = w.unsqueeze(-1).unsqueeze(-1)

        # write processed params to state dict 
        state_dict[prefix+'conv.weight'] = w 
        if b is not None:
            state_dict[prefix+'conv.bias'] = b


class DeConvLayer(Model):
    conv: L.deconv2D
    batch_norm: bool 
    bn: L.BatchNorm
    act: L.Activation

    def __init__(self, size:int, outchn:int, stride:int=1, pad:PadModes='SAME_LEFT', dilation_rate:int=1, activation:int=-1,
                 batch_norm:bool=False, affine:bool=True, usebias:bool=True, groups:int=1):
        super().__init__(size=size, outchn=outchn, stride=stride, pad=pad, dilation_rate=dilation_rate, activation=activation,
                         batch_norm=batch_norm, affine=affine, usebias=usebias, groups=groups)
        
    def initialize(self, size:int, outchn:int, stride:int=1, pad:PadModes='SAME_LEFT', dilation_rate:int=1, activation:int=-1,
                 batch_norm:bool=False, affine:bool=True, usebias:bool=True, groups:int=1):
        self.conv = L.deconv2D(size, outchn, stride, pad, dilation_rate, usebias, groups)
        if batch_norm:
            self.bn = L.BatchNorm(affine=affine)
        self.batch_norm = batch_norm
        self.act = L.Activation(activation)

    def forward(self, x: Tensor) -> Tensor:
        x = self.conv(x)
        if self.batch_norm:
            x = self.bn(x)
        x = self.act(x)
        return x 
    
    def __call__(self, x: Tensor) -> Tensor:
        return super().__call__(x)
    
    def to_torch_module(self):
        conv = self.conv.to_torch()
        setattr(self, 'conv', conv)
        if self.batch_norm:
            bn = self.bn.to_torch()
            setattr(self, 'bn', bn)
        if self.activation==PARAM_RELU:
            relu = nn.ReLU()
            setattr(self, 'act', relu)

    def get_input_quantizer(self) -> QuantizerBase:
        return self.conv.input_quantizer

    def get_weight_quantizer(self) -> QuantizerBase:
        return self.conv.w_quantizer

    def set_input_quantizer(self, quantizer: QuantizerBase):
        self.conv.set_input_quantizer(quantizer)

    def _load_from_state_dict2(self, state_dict, prefix):
        def _load_weight(k):
            if not k in state_dict:
                raise Exception(f'Attenpt to find [{k}] but only exist [{state_dict.keys()}]. Cannot find weight in checkpoint for layer: [{prefix}]')
            return state_dict.pop(k)
        def _load_bias(k):
            if self.conv.usebias:
                try:
                    b = state_dict.pop(k)
                except:
                    raise Exception(f'Attenpt to find [{k}] but only exist [{state_dict.keys()}] Bias is set for layer [{prefix}] but not found in checkpoint.')
            else:
                b = None
            return b 
        
        if not self._is_built:
            logger.warning(f'Layer: [{prefix}] is not built. This layer was not used in dummy forward. Skipping loading state_dict.')
            return 

        # get names for params
        if prefix+'conv.weight' in state_dict:
            # normal load 
            w = prefix + 'conv.weight'
            b = prefix + 'conv.bias'
        elif self.get_flag('from_torch'):
            w = prefix + 'weight'
            b = prefix + 'bias'
        else:
            raise KeyError(f'Cannot find weight in checkpoint for layer: [{prefix}]')

        # laod weight and bias 
        w = _load_weight(w)
        b = _load_bias(b)

        # write processed params to state dict 
        state_dict[prefix+'conv.weight'] = w 
        if b is not None:
            state_dict[prefix+'conv.bias'] = b


class ConvLayer1D(Model):
    conv: L.conv1D
    batch_norm: bool
    bn: L.BatchNorm
    act: L.Activation

    def __init__(self, size: int, outchn:int, stride:int=1, pad:PadModes='SAME_LEFT', dilation_rate:int=1, activation:int=-1,
                 batch_norm:bool=False, affine:bool=True, usebias:bool=True, groups:int=1):
        super().__init__(size=size, outchn=outchn, stride=stride, pad=pad, dilation_rate=dilation_rate, activation=activation,
                         batch_norm=batch_norm, affine=affine, usebias=usebias, groups=groups)
        
    def initialize(self, size: int, outchn:int, stride:int=1, pad:PadModes='SAME_LEFT', dilation_rate:int=1, activation:int=-1,
                 batch_norm:bool=False, affine:bool=True, usebias:bool=True, groups:int=1):
        self.conv = L.conv1D(size, outchn, stride, pad, dilation_rate, usebias, groups)
        if batch_norm:
            self.bn = L.BatchNorm(affine=affine)
        self.batch_norm = batch_norm
        self.act = L.Activation(activation)

    def forward(self, x: Tensor) -> Tensor:
        x = self.conv(x)
        if self.batch_norm:
            x = self.bn(x)
        x = self.act(x)
        return x 
    
    def __call__(self, x: Tensor) -> Tensor:
        return super().__call__(x)
    
    def to_torch_module(self):
        conv = self.conv.to_torch()
        setattr(self, 'conv', conv)
        if self.batch_norm:
            bn = self.bn.to_torch()
            setattr(self, 'bn', bn)
        if self.activation==PARAM_RELU:
            relu = nn.ReLU()
            setattr(self, 'act', relu)

    def get_input_quantizer(self) -> QuantizerBase:
        return self.conv.input_quantizer

    def get_weight_quantizer(self) -> QuantizerBase:
        return self.conv.w_quantizer

    def set_input_quantizer(self, quantizer: QuantizerBase):
        self.conv.set_input_quantizer(quantizer)
    
    def _load_from_state_dict2(self, state_dict, prefix):
        def _load_weight(k):
            if not k in state_dict:
                raise KeyError(f'Attenpt to find [{k}] but only exist [{state_dict.keys()}] Cannot find weight in checkpoint for layer: [{prefix}]')
            return state_dict.pop(k)
        def _load_bias(k):
            if self.conv.usebias:
                try:
                    b = state_dict.pop(k)
                except:
                    raise KeyError(f'Attenpt to find [{k}] but only exist [{state_dict.keys()}] Bias is set for layer [{prefix}] but not found in checkpoint.')
            else:
                b = None
            return b 
        
        if not self._is_built:
            logger.warning(f'Layer: [{prefix}] is not built. This layer was not used in dummy forward. Skipping loading state_dict.')
            return 

        # get names for params
        if prefix+'conv.weight' in state_dict:
            # normal load 
            w = prefix + 'conv.weight'
            b = prefix + 'conv.bias'
        elif self.get_flag('from_torch'):
            w = prefix + 'weight'
            b = prefix + 'bias'
        else:
            raise KeyError(f'Cannot find weight in checkpoint for layer: [{prefix}]')

        # laod weight and bias 
        w = _load_weight(w)
        b = _load_bias(b)

        # write processed params to state dict 
        state_dict[prefix+'conv.weight'] = w 
        if b is not None:
            state_dict[prefix+'conv.bias'] = b


class ConvLayer3D(Model):
    conv: L.conv3D
    batch_norm: bool
    bn: L.BatchNorm
    act: L.Activation

    def __init__(self, size: TypeKSize3D, outchn:int, stride:TypeKSize3D=1, pad:PadModes='SAME_LEFT', dilation_rate:int=1, activation:int=-1,
                 batch_norm:bool=False, affine:bool=True, usebias:bool=True, groups:int=1):
        super().__init__(size=size, outchn=outchn, stride=stride, pad=pad, dilation_rate=dilation_rate, activation=activation,
                         batch_norm=batch_norm, affine=affine, usebias=usebias, groups=groups)
        
    def initialize(self, size: TypeKSize3D, outchn:int, stride:TypeKSize3D=1, pad:PadModes='SAME_LEFT', dilation_rate:int=1, activation:int=-1,
                 batch_norm:bool=False, affine:bool=True, usebias:bool=True, groups:int=1):
        self.conv = L.conv3D(size, outchn, stride, pad, dilation_rate, usebias, groups)
        if batch_norm:
            self.bn = L.BatchNorm(affine=affine)
        self.batch_norm = batch_norm
        self.act = L.Activation(activation)

    def forward(self, x: Tensor) -> Tensor:
        x = self.conv(x)
        if self.batch_norm:
            x = self.bn(x)
        x = self.act(x)
        return x 
    
    def __call__(self, x: Tensor) -> Tensor:
        return super().__call__(x)
    
    def to_torch_module(self):
        conv = self.conv.to_torch()
        setattr(self, 'conv', conv)
        if self.batch_norm:
            bn = self.bn.to_torch()
            setattr(self, 'bn', bn)
        if self.activation==PARAM_RELU:
            relu = nn.ReLU()
            setattr(self, 'act', relu)

    def get_input_quantizer(self) -> QuantizerBase:
        return self.conv.input_quantizer

    def get_weight_quantizer(self) -> QuantizerBase:
        return self.conv.w_quantizer

    def set_input_quantizer(self, quantizer: QuantizerBase):
        self.conv.set_input_quantizer(quantizer)
    
    def _load_from_state_dict2(self, state_dict, prefix):
        def _load_weight(k):
            if not k in state_dict:
                raise KeyError(f'Attenpt to find [{k}] but only exist [{state_dict.keys()}] Cannot find weight in checkpoint for layer: [{prefix}]')
            return state_dict.pop(k)
        def _load_bias(k):
            if self.conv.usebias:
                try:
                    b = state_dict.pop(k)
                except:
                    raise KeyError(f'Attenpt to find [{k}] but only exist [{state_dict.keys()}] Bias is set for layer [{prefix}] but not found in checkpoint.')
            else:
                b = None
            return b 
        
        if not self._is_built:
            logger.warning(f'Layer: [{prefix}] is not built. This layer was not used in dummy forward. Skipping loading state_dict.')
            return 

        # get names for params
        if prefix+'conv.weight' in state_dict:
            # normal load 
            w = prefix + 'conv.weight'
            b = prefix + 'conv.bias'
        elif self.get_flag('from_torch'):
            w = prefix + 'weight'
            b = prefix + 'bias'
        else:
            raise KeyError(f'Cannot find weight in checkpoint for layer: [{prefix}]')
        
        # laod weight and bias 
        w = _load_weight(w)
        b = _load_bias(b)

        # write processed params to state dict 
        state_dict[prefix+'conv.weight'] = w 
        if b is not None:
            state_dict[prefix+'conv.bias'] = b




import torch.nn as nn
from torch import Tensor

from .. import Layers as L
from ..Base import Model
from ..Consts.Activate import *
from ..Consts.Types import *
from ..Quant import QuantizerBase


class Dense(Model):
    fc: L.fcLayer
    bn: L.BatchNorm
    batch_norm: bool 
    act: L.Activation

    def __init__(self, outsize: int, batch_norm:bool=False, affine:bool=True, activation:int=-1, usebias:bool=True, norm:bool=False):
        super().__init__(outsize=outsize, batch_norm=batch_norm, affine=affine, activation=activation, usebias=usebias, norm=norm)

    def initialize(self, outsize, batch_norm=False, affine=True, activation=-1 , usebias=True, norm=False):
        self.fc = L.fcLayer(outsize, usebias, norm)
        self.batch_norm = batch_norm
        self.act = L.Activation(activation)
        if batch_norm:
            self.bn = L.BatchNorm(affine=affine)

    def forward(self, x: Tensor) -> Tensor:
        x = self.fc(x)
        if self.batch_norm:
            assert len(x.shape)==2, f'Batch norm in dense layers only accepts num_dims=2, but got [{len(x.shape)}]'
            x = self.bn(x)
        x = self.act(x)
        return x 
    
    def __call__(self, x: Tensor) -> Tensor:
        return super().__call__(x)

    def get_input_quantizer(self) -> QuantizerBase:
        return self.fc.input_quantizer

    def get_weight_quantizer(self) -> QuantizerBase:
        return self.fc.w_quantizer

    def set_input_quantizer(self, quantizer: QuantizerBase):
        self.fc.set_input_quantizer(quantizer)

    def to_torch_module(self):
        fc = self.fc.to_torch()
        setattr(self, 'fc', fc)
        if self.batch_norm:
            bn = self.bn.to_torch()
            setattr(self, 'bn', bn)
        if self.activation==PARAM_RELU:
            relu = nn.ReLU()
            setattr(self, 'act', relu)

    def _load_from_state_dict2(self, state_dict, prefix):
        if prefix+'fc.weight' in state_dict:
            return 
        if self.get_flag('from_torch'):
            w = state_dict.pop(prefix + 'weight')
            if self.fc.usebias:
                try:
                    b = state_dict.pop(prefix + 'bias')
                except:
                    raise KeyError(f'Bias is set for layer [{prefix}] but not found in checkpoint.')
            else:
                b = None

            state_dict[prefix+'fc.weight'] = w
            if b is not None: 
                state_dict[prefix+'fc.bias'] = b




import os
from distutils.version import LooseVersion

import torch
import torch.nn as nn
from loguru import logger

from ..Base import Model


class Saver():
    model: Model 

    def __init__(self, module: Model):
        self.model = module

    def _get_checkpoint(self, path: str) -> str:
        ckpt = os.path.join(path, 'checkpoint')
        if os.path.exists(ckpt):
            fname = open(ckpt).readline().strip()
            return os.path.join(path, fname)
        else:
            return ''

    def restore(self, path: str, strict: bool=True):
        logger.info('Trying to load from: %s'%path)
        device = torch.device('cpu')
        if path[-4:] == '.pth':
            if not os.path.exists(path):
                logger.warning('Path: %s does not exsist. No restoration will be performed.'%path)
            elif isinstance(self.model, (nn.DataParallel, nn.parallel.DistributedDataParallel)):
                state_dict = torch.load(path, map_location=device)
                self.model.module.load_state_dict(state_dict, strict=strict)
                logger.info('Model loaded from: %s'%path)
            else:
                state_dict = torch.load(path, map_location=device)
                self.model.load_state_dict(state_dict, strict=strict)
                logger.info('Model loaded from: %s'%path)
        else:
            path = self._get_checkpoint(path)
            if path:
                if isinstance(self.model, (nn.DataParallel, nn.parallel.DistributedDataParallel)):
                    state_dict = torch.load(path, map_location=device)
                    self.model.module.load_state_dict(state_dict, strict=strict)
                else:
                    state_dict = torch.load(path, map_location=device)
                    self.model.load_state_dict(state_dict, strict=strict)
                logger.info('Model loaded from: %s'%path)
            else:
                logger.warning('No checkpoint found. No restoration will be performed.')

    def save(self, path: str):
        # To make it compatible with older pytorch 
        directory = os.path.dirname(path)
        os.makedirs(directory, exist_ok=True)
        if isinstance(self.model, (nn.DataParallel, nn.parallel.DistributedDataParallel)):
            if LooseVersion(torch.__version__)>=LooseVersion('1.6.0'):
                torch.save(self.model.module.state_dict(), path, _use_new_zipfile_serialization=False)
            else:
                torch.save(self.model.module.state_dict(), path)
        else:
            if LooseVersion(torch.__version__)>=LooseVersion('1.6.0'):
                torch.save(self.model.state_dict(), path, _use_new_zipfile_serialization=False)
            else:
                torch.save(self.model.state_dict(), path)
        logger.info('Model saved to: %s'%path)
        ckpt = open(directory + '/checkpoint', 'w')
        ckpt.write(os.path.basename(path))
        ckpt.close()




# activation const
# some values are no longer supported 

PARAM_RELU = 0
PARAM_LRELU = 1
PARAM_ELU = 2
PARAM_TANH = 3
PARAM_SIGMOID = 6
PARAM_PRELU = 8
PARAM_PRELU1 = 9
PARAM_GELU = 10
PARAM_SILU = 11



import torch
from torch import Tensor
from typing import Tuple, List 

from .. import Layers as L

from.Conv import ConvLayer
from ..Base import Model
from ..Consts.Types import *
from .Dense import Dense


class LSTMCell(Model):
    def __init__(self, outdim: int, usebias: bool=False):
        super().__init__(outdim=outdim, usebias=usebias)

    def initialize(self, outdim: int, usebias: bool=False):
        self.F = Dense(outdim, usebias=usebias, norm=False)
        self.O = Dense(outdim, usebias=usebias, norm=False)
        self.I = Dense(outdim, usebias=usebias, norm=False)
        self.C = Dense(outdim, usebias=usebias, norm=False)

        self.hF = Dense(outdim, usebias=usebias, norm=False)
        self.hO = Dense(outdim, usebias=usebias, norm=False)
        self.hI = Dense(outdim, usebias=usebias, norm=False)
        self.hC = Dense(outdim, usebias=usebias, norm=False)

    def forward(self, x: Tensor, h: Tensor, c_prev: Tensor) -> Tuple[Tensor, Tensor]:
        f = self.F(x) + self.hF(h)
        o = self.O(x) + self.hO(h)
        i = self.I(x) + self.hI(h)
        c = self.C(x) + self.hC(h)

        f_ = torch.sigmoid(f)
        c_ = torch.tanh(c) * torch.sigmoid(i)
        o_ = torch.sigmoid(o)

        next_c = c_prev * f_ + c_ 
        next_h = o_ * torch.tanh(next_c)
        return next_h, next_c
    
    def __call__(self, x: Tensor, h: Tensor, c_prev: Tensor) -> Tuple[Tensor, Tensor]:
        return super().__call__(x, h, c_prev)


class ConvLSTM(Model):
    def __init__(self, chn: int, usebias: bool=False):
        super().__init__(chn=chn, usebias=usebias)

    def initialize(self, chn: int, usebias: bool=False):
        self.gx = ConvLayer(3, chn, usebias=usebias)
        self.gh = ConvLayer(3, chn, usebias=usebias)
        self.fx = ConvLayer(3, chn, usebias=usebias)
        self.fh = ConvLayer(3, chn, usebias=usebias)
        self.ox = ConvLayer(3, chn, usebias=usebias)
        self.oh = ConvLayer(3, chn, usebias=usebias)
        self.ix = ConvLayer(3, chn, usebias=usebias)
        self.ih = ConvLayer(3, chn, usebias=usebias)

    def forward(self, x: Tensor, c: Tensor, h: Tensor) -> Tuple[Tensor, Tensor]:
        gx = self.gx(x)
        gh = self.gh(h)

        ox = self.ox(x)
        oh = self.oh(h)

        fx = self.fx(x)
        fh = self.fh(h)

        ix = self.ix(x)
        ih = self.ih(h)

        g = torch.tanh(gx + gh)
        o = torch.sigmoid(ox + oh)
        i = torch.sigmoid(ix + ih)
        f = torch.sigmoid(fx + fh)

        cell = f*c + i*g 
        h = o * torch.tanh(cell)
        return cell, h 
    
    def __call__(self, x: Tensor, c: Tensor, h: Tensor) -> Tuple[Tensor, Tensor]:
        return super().__call__(x, c, h)


class AdaptConv3(Model):
    # Originally proposed in DEKR
    regular_matrix: Tensor
    transform_conv: ConvLayer
    translation_conv: ConvLayer
    deform_conv: L.DeformConv2D
    batch_norm: bool
    bn: L.BatchNorm
    act: L.Activation

    def __init__(self, outchn: int, stride: int=1, pad: PadModes='SAME_LEFT', dilation_rate: int=1, batch_norm: bool=False, activation: int=-1, usebias: bool=True):
        super().__init__(outchn=outchn, stride=stride, pad=pad, dilation_rate=dilation_rate, batch_norm=batch_norm, activation=activation, usebias=usebias)

    def initialize(self, outchn: int, stride: int=1, pad: PadModes='SAME_LEFT', dilation_rate: int=1, batch_norm: bool=False, activation: int=-1, usebias: bool=True):
        regular_matrix = torch.tensor([[-1, -1, -1, 0, 0, 0, 1, 1, 1],\
            [-1, 0, 1, -1 ,0 ,1 ,-1, 0, 1]])
        # register to buffer that it can be mangaed by cuda or cpu
        self.register_buffer('regular_matrix', regular_matrix.float())
        self.transform_conv = ConvLayer(3, 4)
        self.translation_conv = ConvLayer(3, 2)
        self.deform_conv = L.DeformConv2D(3, outchn, stride, pad, dilation_rate, usebias)
        self.batch_norm = batch_norm
        if batch_norm:
            self.bn = L.BatchNorm()
        self.act = L.Activation(activation)

    def forward(self, x: Tensor) -> Tensor:
        N, C, H, W = x.shape
        trans_mtx = self.transform_conv(x)
        trans_mtx = trans_mtx.permute(0,2,3,1).reshape((N*H*W,2,2))
        offset = torch.matmul(trans_mtx, self.regular_matrix)
        offset = offset-self.regular_matrix
        offset = offset.transpose(1,2).reshape((N,H,W,18)).permute(0,3,1,2)

        translation = self.translation_conv(x)
        offset[:,0::2,:,:] += translation[:,0:1,:,:]
        offset[:,1::2,:,:] += translation[:,1:2,:,:]

        out = self.deform_conv(x, offset)
        
        if self.batch_norm:
            out = self.bn(out)
        
        if self.activation!=-1:
            out = self.act(out)
        return out 
    
    def __call__(self, x: Tensor) -> Tensor:
        return super().__call__(x)



from typing import Union

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.nn.init as init
from torch import Tensor
from torch.nn.parameter import Parameter

from ..Base import Model


class Transpose_last(nn.Module):
    dim:int 

    def __init__(self, dim:int):
        self.dim = dim 

    def forward(self, x:Tensor) -> Tensor:
        if self.dim==-1:
            return x 
        return x.transpose(-1, self.dim)


class BatchNorm(Model):
    eps: Union[float, None]
    momentum: float 
    affine: bool 
    track_running_stats: bool 
    num_batches_tracked: Tensor
    running_mean: Tensor 
    running_var: Tensor
    n_dims: int

    def __init__(self, eps:Union[float, None]=None, momentum:float=0.01, affine:bool=True, track_running_stats:bool=True):
        super().__init__(eps=eps, momentum=momentum, affine=affine, track_running_stats=track_running_stats)

    def initialize(self, eps:Union[float, None]=None, momentum:float=0.01, affine:bool=True, track_running_stats:bool=True):
        self.eps = eps
        self.momentum = momentum
        self.affine = affine
        self.track_running_stats = track_running_stats
        
    def build(self, *inputs):
        num_features = inputs[0].shape[1]
        self.n_dims = len(inputs[0].shape)
        self.num_features = num_features
        if self.affine:
            self.weight = Parameter(torch.Tensor(num_features))
            self.bias = Parameter(torch.Tensor(num_features))
        else:
            self.register_parameter('weight', None)
            self.register_parameter('bias', None)
        if self.track_running_stats:
            self.register_buffer('running_mean', torch.zeros(num_features))
            self.register_buffer('running_var', torch.ones(num_features))
            self.register_buffer('num_batches_tracked', torch.tensor(0, dtype=torch.long))
        else:
            self.register_parameter('running_mean', None)
            self.register_parameter('running_var', None)
            self.register_parameter('num_batches_tracked', None)
        self.reset_parameters()
        eps = self.get_flag('bn_eps')
        if eps is not None:
            self.eps = eps 
        if self.eps is None:
            self.eps = 1e-5

    def reset_running_stats(self):
        if self.track_running_stats:
            self.running_mean.zero_() 
            self.running_var.fill_(1) 
            self.num_batches_tracked.zero_()

    def reset_parameters(self):
        self.reset_running_stats()
        if self.affine:
            init.ones_(self.weight)
            init.zeros_(self.bias)

    def forward(self, input):
        if self.momentum is None:
            exponential_average_factor = 0.0
        else:
            exponential_average_factor = self.momentum

        if self.training and self.track_running_stats:
            # TODO: if statement only here to tell the jit to skip emitting this when it is None
            if self.num_batches_tracked is not None:
                self.num_batches_tracked += 1
                if self.momentum is None:  # use cumulative moving average
                    exponential_average_factor = 1.0 / float(self.num_batches_tracked)
                else:  # use exponential moving average
                    exponential_average_factor = self.momentum

        if self.eps is not None:
            result =  F.batch_norm(
                input, self.running_mean, self.running_var, self.weight, self.bias,
                self.training or not self.track_running_stats,
                exponential_average_factor, self.eps)
        else:
            raise Exception('eps is not set for batch norm layer.')
        return result
    
    def __call__(self, x:Tensor) -> Tensor:
        return super().__call__(x)

    def to_torch(self) -> nn.modules.batchnorm._BatchNorm:
        if self.eps is not None:
            if self.n_dims==3 or self.n_dims==2:
                bn = nn.BatchNorm1d(self.num_features, self.eps, self.momentum, self.affine, self.track_running_stats)
            elif self.n_dims==4:
                bn = nn.BatchNorm2d(self.num_features, self.eps, self.momentum, self.affine, self.track_running_stats)
            elif self.n_dims==5:
                bn = nn.BatchNorm3d(self.num_features, self.eps, self.momentum, self.affine, self.track_running_stats)
            else:
                raise Exception('Number of dims in batch norm should be either 3,4,5')
            if self.affine:
                bn.weight.data[:] = self.weight.data[:]
                bn.bias.data[:] = self.bias.data[:]
            if (bn.running_mean is not None) and (bn.running_var is not None):
                bn.running_mean.data[:] = self.running_mean.data[:]
                bn.running_var.data[:] = self.running_var.data[:]
            return bn 
        else:
            raise Exception('eps is not set for batch norm layer')


class LayerNorm(Model):
    dim: int 
    affine: bool 
    eps: Union[float, None]
    weight: Parameter
    bias: Parameter
    chn: int

    def initialize(self, dim=-1, affine=True, eps=None):
        self.dim = dim
        self.affine = affine
        self.eps = eps 

    def build(self, *inputs):
        shape = inputs[0].shape
        self.chn = shape[self.dim]
        if self.affine:
            self.weight = Parameter(torch.Tensor(shape[self.dim]))
            self.bias = Parameter(torch.Tensor(shape[self.dim]))
        self.reset_params()
        eps = self.get_flag('ln_eps')
        if eps is not None:
            self.eps = eps 
        if self.eps is None:
            self.eps = 1e-5

    def reset_params(self):
        if self.affine:
            init.ones_(self.weight)
            init.zeros_(self.bias)

    def forward(self, x: Tensor) -> Tensor:
        if self.dim!=-1:
            x = x.transpose(self.dim, -1)
        mean = torch.mean(x, dim=-1, keepdim=True)
        var = torch.var(x, dim=-1, correction=0, keepdim=True)
        if self.affine:
            out = self.weight * (x - mean) / torch.sqrt(var + self.eps) + self.bias
        else:
            out = (x - mean) / torch.sqrt(var + self.eps)
        if self.dim!=-1:
            out = out.transpose(self.dim, -1)
        return out 
        
    def __call__(self, x:Tensor) -> Tensor:
        return super().__call__(x)

    def to_torch(self) -> nn.Sequential:
        mod = nn.Sequential(Transpose_last(self.dim), nn.LayerNorm(normalized_shape=self.chn), Transpose_last(self.dim))
        return mod 




from .QCalibrator import LayerCalibrator
from .Quantizers import QQuantizers
from ..Base import Model
from torch import Tensor 


class QMatmul(Model):
    def __init__(self, swap_ab: bool=False):
        super().__init__(swap_ab)

    def initialize(self, swap_ab: bool=False):
        self.swap_ab = swap_ab

    def build(self, *args, **kwargs):
        if self._quant:
            bit_type = self.get_flag('QActBit')
            if bit_type is None:
                bit_type = 'int8'

            if self.get_flag('LayerwiseQuant'):
                self.a_quantizer = QQuantizers['uniform'](zero_offset=True, observer='placeholder')
                self.b_quantizer = QQuantizers['uniform'](zero_offset=True, observer='placeholder')
                self.calibrator = LayerCalibrator([self.a_quantizer, self.b_quantizer], self.forward)
                self.register_forward_hook(self.calibrator.layer_hook)
            else:
                obs_type = self.get_flag('QActObserver')
                if obs_type is None:
                    obs_type = 'minmax'
                self.a_quantizer = QQuantizers['uniform'](zero_offset=True, observer=obs_type)
                self.b_quantizer = QQuantizers['uniform'](zero_offset=True, observer=obs_type)

    def forward(self, a, b):
        if self._quant:
            if self.swap_ab:
                a = self.b_quantizer(a)
                b = self.a_quantizer(b)
            else:
                a = self.a_quantizer(a)
                b = self.b_quantizer(b)
        return a @ b 
    
    def __call__(self, a: Tensor, b: Tensor) -> Tensor:
        return super().__call__(a, b)


class QAdd(Model):
    def __init__(self, swap_ab: bool=False, quant_left: bool=True, quant_right: bool=True):
        super().__init__(swap_ab, quant_left, quant_right)

    def initialize(self, swap_ab: bool=False, quant_left: bool=True, quant_right: bool=True) -> None:
        self.swap_ab = swap_ab
        self.quant_left = quant_left
        self.quant_right = quant_right

    def build(self, *args, **kwargs):
        if self._quant:
            bit_type = self.get_flag('QActBit')
            if bit_type is None:
                bit_type = 'int8'

            if self.get_flag('LayerwiseQuant'):
                quantizers = []
                if (self.quant_left and not self.swap_ab) or (self.quant_right and self.swap_ab):
                    self.a_quantizer = QQuantizers['uniform'](zero_offset=False, observer='placeholder')
                    quantizers.append(self.a_quantizer)
                if (self.quant_right and not self.swap_ab) or (self.quant_left and self.swap_ab):
                    self.b_quantizer = QQuantizers['uniform'](zero_offset=False, observer='placeholder')
                    quantizers.append(self.b_quantizer)
                self.calibrator = LayerCalibrator(quantizers, self.forward)
                self.register_forward_hook(self.calibrator.layer_hook)
            else:
                obs_type = self.get_flag('QActObserver')
                if obs_type is None:
                    obs_type = 'minmax'
                if (self.quant_left and not self.swap_ab) or (self.quant_right and self.swap_ab):
                    self.a_quantizer = QQuantizers['uniform'](zero_offset=False, observer=obs_type)
                if (self.quant_right and not self.swap_ab) or (self.quant_left and self.swap_ab):
                    self.b_quantizer = QQuantizers['uniform'](zero_offset=False, observer=obs_type)

    def forward(self, a, b):
        if self._quant:
            if self.swap_ab:
                if self.quant_left:
                    a = self.b_quantizer(a)
                if self.quant_right:
                    b = self.a_quantizer(b)
            else:
                if self.quant_left:
                    a = self.a_quantizer(a)
                if self.quant_right:
                    b = self.b_quantizer(b)
        return a + b 

    def __call__(self, a: Tensor, b: Tensor) -> Tensor:
        return super().__call__(a, b)
    



import sys 
from rich.progress import (BarColumn, MofNCompleteColumn, Progress,
                           ProgressColumn, TextColumn, TimeRemainingColumn)
from rich.text import Text


# progressbar utils 
class SpeedColumn(ProgressColumn):
    def render(self, task):
        speed = task.finished_speed or task.speed
        if speed is None:
            return Text("?", style="progress.data.speed")
        else:
            return Text(f"{speed:.2f} it/s", style="progress.data.speed")

def progress_bar(width: int=40, disable: bool=False) -> Progress:
    # is_debug_mode = False
    # gettrace = getattr(sys, 'gettrace', None)
    # if gettrace is not None:
    #     if gettrace() is not None:
    #         is_debug_mode = True

    prog = Progress(
        TextColumn('[progress.description]{task.description}'), 
        BarColumn(finished_style='green', bar_width=width), 
        MofNCompleteColumn(), 
        TimeRemainingColumn(elapsed_when_finished=True), 
        SpeedColumn(),
        disable = disable,
        )
    return prog



from . import Base
from . import Layers as L
from . import Modules
from . import Quant as Qnt
from . import Tools, Utils
from .Consts.Activate import * 


## Base model 
Model = Base.Model

## Modules 
ConvLayer = Modules.ConvLayer
ConvLayer1D = Modules.ConvLayer1D
ConvLayer3D = Modules.ConvLayer3D
DeConvLayer = Modules.DeConvLayer
Dense = Modules.Dense
LSTMCell = Modules.LSTMCell
ConvLSTM = Modules.ConvLSTM
AdaptConv3 = Modules.AdaptConv3

## Layers 
Activation = L.Activation
DeformConv2D = L.DeformConv2D
BatchNorm = L.BatchNorm
LayerNorm = L.LayerNorm
MaxPool2D = L.MaxPool2d
AvgPool2D = L.AvgPool2d
BilinearUpSample = L.BilinearUpSample
NNUpSample = L.NNUpSample

## Quantization related 
QAct = Qnt.QAct
QMatmul = Qnt.QMatmul
QAdd = Qnt.QAdd

## Utils 
Saver = Utils.Saver
init_model = Utils.init_model
to_standard_torch = Utils.to_standard_torch
inspect_quant_params = Utils.inspect_quant_params

## Tools 
BBox = Tools.BBox
# ## The following is suggested to be accessed by importing TorchSUL.Tools 
# Path = Tools.Path
# VideoSaver = Tools.VideoSaver
# check_fps = Tools.check_fps
# check_frame_num = Tools.check_frame_num
# combine_audio = Tools.combine_audio
# compress_video = Tools.compress_video
# extract_frames = Tools.extract_frames
# progress_bar = Tools.progress_bar
# Visualizer = Tools.Visualizer
# VisualizerBase = Visualizer.VisualizerBase
# start_visualize = Visualizer.start_visualize




from typing import Union, List, Tuple
from typing_extensions import Literal


## Layer param type declaration
PadModes = Literal['SAME_LEFT', 'VALID']
TypeKSize = Union[int, List[int], Tuple[int,...]]
TypeKSize2D = Union[int, List[int], Tuple[int,int]]
TypeKSize3D = Union[int, List[int], Tuple[int,int,int]]

## Quant type declarations
QuantModes = Literal['layer_wise', 'channel_wise']
QBitTypes = Literal['uint8', 'int8', 'int16']
QObserverTypes = Literal['minmax', 'percentile', 'omse', 'placeholder']
QuantizerTypes = Literal['uniform']

## Image-video related 
BasicImageTypes = Literal['jpg', 'png', 'jpeg']

# These are pre-set flags for SULModel 
PRESET_FLAGS = Literal['save_tensor', 'loose_load', 'from_torch', 'fc2conv', 'QActBit', 'QActObserver',\
                        'conv_init_mode', 'fc_init_mode', 'bn_eps', 'ln_eps', 'dump_onnx', 'LayerwiseQuant', 'CalibScoreFunc']



from typing import Union

from torch import Tensor

from ..Base import Model
from ..Consts.Types import *
from .Quantizers import QQuantizers, QuantizerBase


class QAct(Model):
    bit_type: Union[QBitTypes, None]
    mode: QuantModes
    observer_str: Union[QObserverTypes, None]
    is_weight: bool
    zero_offset: bool

    def __init__(self, zero_offset: bool=False, mode: QuantModes='layer_wise', \
                    observer: Union[QObserverTypes,None]=None, bit_type: Union[QBitTypes, None]=None, is_weight: bool=False):
        super().__init__(zero_offset=zero_offset, mode=mode, observer=observer, bit_type=bit_type, is_weight=is_weight)           # make typechecker happy
        
    def initialize(self, zero_offset: bool=False, mode: QuantModes='layer_wise', \
                    observer: Union[QObserverTypes,None]=None, bit_type: Union[QBitTypes, None]=None, is_weight: bool=False):
        self.mode = mode 
        self.zero_offset = zero_offset
        self.observer_str = observer
        self.bit_type = bit_type
        self.is_weight = is_weight

    def build(self, x):
        if self._quant:
            bit_type: Union[QBitTypes, None] = self.get_flag('QActBit')
            if bit_type is None:
                bit_type = 'int8'
            if self.bit_type is not None:
                bit_type = self.bit_type

            obs_type = self.get_flag('QActObserver')
            if obs_type is None:
                obs_type = 'minmax'
            if self.observer_str is not None:
                obs_type = self.observer_str
            self.quantizer = QQuantizers['uniform'](bit_type=bit_type, zero_offset=self.zero_offset, mode=self.mode, observer=obs_type, is_weight=self.is_weight)

    def forward(self, x: Tensor) -> Tensor:
        if self._quant:
            x = self.quantizer(x)
        return x 

    def __call__(self, x: Tensor) -> Tensor:
        return super().__call__(x)
    
    def get_quantizer(self) -> QuantizerBase:
        return self.quantizer
    
    def set_quantizer(self, quantizer: QuantizerBase):
        if quantizer is not None:
            assert isinstance(quantizer, QuantizerBase), 'Must assign quantizer'
            self.quantizer = quantizer
    



import torch
from torch.autograd import Function
from torch.autograd.function import once_differentiable


##### START: Quant aware training function
class QATFunc(Function):
    # quant-aware training function 
    @staticmethod 
    def forward(ctx, x, scale, zero_point, Qn, Qp, zero_offset=False, mode='layer_wise', dim=-1):
        ctx.Qn = Qn 
        ctx.Qp = Qp 
        ctx.mode = mode 
        ctx.dim = dim 
        ctx.zero_offset = zero_offset

        if dim!=-1 and mode=='channel_wise':
            # make everything run at last dim, no need manually reshape 
            x = x.transpose(-1, dim)

        zero_point = zero_point.round()
        x_back = x 
        x = x / scale + zero_point
        x = x.round().clamp(Qn, Qp)
        ctx.save_for_backward(x_back, x, scale, zero_point)     # save here to save computation for backward

        x = (x - zero_point) * scale 
        if dim!=-1 and mode=='channel_wise':
            x = x.transpose(-1, dim)
        return x

    @staticmethod
    @once_differentiable
    def backward(ctx, out_grad):
        out_grad = out_grad.contiguous()
        x, x1, scale, zero_point = ctx.saved_tensors

        # compute ds 
        ds = x1.clone()
        ds -= zero_point
        idx = (x1>ctx.Qn) & (x1<ctx.Qp)
        if ctx.mode=='channel_wise':
            coef = torch.zeros_like(x)
            coef[idx] += x[idx]
            ds -= coef / scale
            ds *= out_grad.transpose(ctx.dim, -1)
        else:
            ds[idx] -= x[idx]/scale
            ds *= out_grad

        # compute db 
        db = torch.zeros_like(x1)
        if not ctx.zero_offset:
            idx = (x1<=ctx.Qn) | (x1>=ctx.Qp)
            if ctx.mode=='channel_wise':
                db[idx] += 1
                db *= -scale
                db *= out_grad.transpose(ctx.dim, -1)
            else:
                db[idx] -= scale
                db *= out_grad

        # compute dx 
        dx = out_grad.clone()
        if ctx.mode=='channel_wise':
            x1 = x1.transpose(ctx.dim, -1)
        dx[(x1<=ctx.Qn) | (x1>=ctx.Qp)] = 0

        if ctx.mode=='channel_wise':
            # use mean here 
            ds = ds.flatten(0,-2).sum(dim=0)
            db = db.flatten(0,-2).sum(dim=0)
        else:
            ds = ds.sum()
            db = db.sum()

        return dx.contiguous(), ds.contiguous(), db.contiguous(), None, None, None, None, None

    @staticmethod
    def symbolic(g, x, scale, zero_point, Qn, Qp, zero_offset=False, mode='layer_wise', dim=-1):
        o1 = g.op('QuantizeLinear', x, scale, zero_point, axis_i=dim)
        return g.op('DequantizeLinear', o1, scale, zero_point, axis_i=dim)
##### END: Quant aware training function




from abc import ABC, abstractmethod

from torch import Tensor
from typing import Dict, Type

from ..Base import Model
from ..Consts.Types import *
from .Observers import QObservers
from .QTypes import QTYPES
from .QATop import QATFunc


##### START: Quantizer 
class QuantizerBase(Model, ABC):
    def __init__(self, bit_type: QBitTypes='int8', observer: QObserverTypes='minmax', zero_offset=False, mode: QuantModes='layer_wise', is_weight=False):
        super().__init__(bit_type=bit_type, observer=observer, zero_offset=zero_offset, mode=mode, is_weight=is_weight)              # make typechecker happy

    def initialize(self, bit_type: QBitTypes='int8', observer: QObserverTypes='minmax', zero_offset=False, mode: QuantModes='layer_wise', is_weight=False):
        self.bit_type = QTYPES[bit_type]
        self.observer = QObservers[observer](self.bit_type, zero_offset, mode, is_weight)
        self.mode = mode 
        self.is_weight = is_weight
        self.zero_offset = zero_offset

    def build(self, x):
        if self.is_weight:
            self.dim = 0
        else:
            if len(x.shape)==4:
                self.dim = 1 
            else:
                self.dim = -1

    @abstractmethod
    def forward(self):
        ...

    def __call__(self, x: Tensor) -> Tensor:
        return super().__call__(x)


class UniformQuantizer(QuantizerBase):
    def forward(self, x: Tensor) -> Tensor:
        x = x.contiguous()
        if self._quant_calibrating:
            x = self.observer(x)
        if self._quant and (self.observer.scale is not None):
            if self.observer.scale.device!=x.device:
                self.observer.to(x.device)
            if self.get_flag('dump_onnx'):
                x = QATFunc.apply(x, self.observer.scale.data.reshape(-1), self.observer.zero_point.data.reshape(-1),     # type: ignore
                      self.bit_type.min_val, self.bit_type.max_val, self.zero_offset, self.mode, self.dim)     # type: ignore
            else:
                x = QATFunc.apply(x, self.observer.scale.contiguous().reshape(-1), self.observer.zero_point.contiguous().reshape(-1),     # type: ignore
                      self.bit_type.min_val, self.bit_type.max_val, self.zero_offset, self.mode, self.dim)     # type: ignore
        return x.contiguous() 


QQuantizers: Dict[Union[QuantizerTypes, str], Type[QuantizerBase]] = {"uniform": UniformQuantizer}
##### END: Quantizer 




import math
from abc import ABC, abstractmethod
from typing import Union, Tuple, List, Dict

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.nn.init as init
from torch import Tensor
from torch.nn.parameter import Parameter

from ..Base import Model
from ..Consts.Types import *
from ..Quant import QQuantizers, QuantizerBase, LayerCalibrator


## Resnet style initialization 
def _resnet_normal(tensor):
    fan_in, fan_out = init._calculate_fan_in_and_fan_out(tensor)
    std = math.sqrt(2.0 / float(fan_out))
    return init._no_grad_normal_(tensor, 0., std)


######  Layers 
class ConvBase(Model, ABC):
    size: TypeKSize
    stride: TypeKSize
    kernel_size = Tuple[int,int,int,int]
    outchn: int 
    pad_mode: PadModes
    dilation_rate: int 
    usebias: bool
    groups: int 
    pad: Union[int, Tuple[int,int]]
    groups: int
    inchn: int
    weight: Parameter
    bias: Union[Parameter, None]
    input_quantizer: QuantizerBase
    w_quantizer: QuantizerBase

    def __init__(self, size:TypeKSize, outchn:int, stride:TypeKSize=1, pad:PadModes='SAME_LEFT', dilation_rate:int=1, usebias:bool=True, groups:int=1):
        super().__init__(size=size, outchn=outchn, stride=stride, pad=pad, dilation_rate=dilation_rate, usebias=usebias, groups=groups)

    def initialize(self, size:TypeKSize, outchn:int, stride:TypeKSize=1, pad:PadModes='SAME_LEFT', dilation_rate:int=1, usebias:bool=True, groups:int=1) -> None:
        self.size = size
        self.outchn = outchn
        self.stride = stride
        self.usebias = usebias
        self.groups = groups
        self.dilation_rate = dilation_rate
        assert (pad in ['VALID','SAME_LEFT'])
        self.pad_mode = pad 
        self.bias = None

    def build(self, *inputs):
        inp = inputs[0]
        self._parse_args(inp.shape)
        self.weight = Parameter(torch.Tensor(*self.kernel_size))
        if self.usebias:
            self.bias = Parameter(torch.Tensor(self.outchn))
        self.reset_params()

        if self._quant:
            bit_type = self.get_flag('QActBit')
            if bit_type is None:
                bit_type = 'int8'
            
            if self.get_flag('LayerwiseQuant'):
                self.input_quantizer = QQuantizers['uniform'](zero_offset=False, bit_type=bit_type, observer='placeholder')
                self.w_quantizer = QQuantizers['uniform'](zero_offset=True, mode='channel_wise', is_weight=True, observer='placeholder')
                self.calibrator = LayerCalibrator([self.input_quantizer, self.w_quantizer], self.forward)
                self.register_forward_hook(self.calibrator.layer_hook)
            else:
                obs_type = self.get_flag('QActObserver')
                if obs_type is None:
                    obs_type = 'minmax'
                self.input_quantizer = QQuantizers['uniform'](zero_offset=False, bit_type=bit_type, observer=obs_type)
                self.w_quantizer = QQuantizers['uniform'](zero_offset=True, mode='channel_wise', is_weight=True)

    @abstractmethod
    def _parse_args(self, shape: List[int]):
        ...

    def set_input_quantizer(self, quantizer: QuantizerBase):
        if quantizer is not None:
            assert isinstance(quantizer, QuantizerBase), 'Must assign quantizer'
            self.input_quantizer = quantizer

    def reset_params(self):
        if self.get_flag('conv_init_mode')=='normal':
            init.normal_(self.weight, std=0.001)
        elif self.get_flag('conv_init_mode')=='resnet':
            _resnet_normal(self.weight)
        else:
            init.kaiming_uniform_(self.weight, a=math.sqrt(5))  # type: ignore
        if self.bias is not None:
            if self.get_flag('conv_init_mode')=='normal':
                init.zeros_(self.bias)
            elif self.get_flag('conv_init_mode')=='resnet':
                fan_in, _ = init._calculate_fan_in_and_fan_out(self.weight)
                bound = 1 / math.sqrt(fan_in)
                init.uniform_(self.bias, -bound, bound)
            else:
                fan_in, _ = init._calculate_fan_in_and_fan_out(self.weight)
                bound = 1 / math.sqrt(fan_in)
                init.uniform_(self.bias, -bound, bound)
    
    def __call__(self, x:Tensor) -> Tensor:
        return super().__call__(x)


class conv2D(ConvBase):
    size: TypeKSize2D
    stride: TypeKSize2D
    def __init__(self, size:TypeKSize2D, outchn:int, stride:TypeKSize2D=1, pad:PadModes='SAME_LEFT', dilation_rate:int=1, usebias:bool=True, groups:int=1):
        super().__init__(size=size, outchn=outchn, stride=stride, pad=pad, dilation_rate=dilation_rate, usebias=usebias, groups=groups)

    def initialize(self, size:TypeKSize2D, outchn:int, stride:TypeKSize2D=1, pad:PadModes='SAME_LEFT', dilation_rate:int=1, usebias:bool=True, groups:int=1) -> None:
        super().initialize(size=size, outchn=outchn, stride=stride, pad=pad, dilation_rate=dilation_rate, usebias=usebias, groups=groups)
    
    def _parse_args(self, input_shape: List[int]):
        inchannel = input_shape[1]
        self.inchn = inchannel
        # parse args
        if isinstance(self.size, list) or isinstance(self.size, tuple):
            if self.pad_mode == 'VALID':
                self.pad = 0
            else:
                self.pad = ((self.size[0]+ (self.dilation_rate-1) * ( self.size[0]-1 ))//2, (self.size[1]+ (self.dilation_rate-1) * (self.size[1]-1))//2)
            self.kernel_size = (self.outchn, inchannel // self.groups, self.size[0], self.size[1])
        elif isinstance(self.size, int):
            if self.pad_mode == 'VALID':
                self.pad = 0
            else:
                self.pad = (self.size + (self.dilation_rate-1) * (self.size-1))//2
            self.kernel_size = (self.outchn, inchannel // self.groups, self.size, self.size)

    def forward(self, x) -> Tensor:
        weight = self.weight
        if self._quant:
            x = self.input_quantizer(x)
            weight = self.w_quantizer(weight)
        return F.conv2d(x, weight, self.bias, self.stride, self.pad, self.dilation_rate, self.groups)

    def to_torch(self) -> nn.Conv2d:
        conv = nn.Conv2d(in_channels = self.inchn, out_channels = self.outchn, kernel_size = self.kernel_size[2:], stride = self.stride,      # type: ignore
                        padding = self.pad, padding_mode = 'zeros', dilation = self.dilation_rate, groups = self.groups, bias = self.usebias)
        conv.weight.data[:] = self.weight.data[:]
        if (self.bias is not None) and (conv.bias is not None):
            conv.bias.data[:] = self.bias.data[:]
        return conv 


class deconv2D(ConvBase):
    size: int
    out_pad: int
    stride: int

    def __init__(self, size:int, outchn:int, stride:int=1, pad:PadModes='SAME_LEFT', dilation_rate:int=1, usebias:bool=True, groups:int=1):
        super().__init__(size=size, outchn=outchn, stride=stride, pad=pad, dilation_rate=dilation_rate, usebias=usebias, groups=groups)

    def initialize(self, size:int, outchn:int, stride:int=1, pad:PadModes='SAME_LEFT', dilation_rate:int=1, usebias:bool=True, groups:int=1) -> None:
        super().initialize(size=size, outchn=outchn, stride=stride, pad=pad, dilation_rate=dilation_rate, usebias=usebias, groups=groups)

    def _parse_args(self, input_shape):
        inchannel = input_shape[1]
        # parse args
        if isinstance(self.size,int):
            if self.pad_mode == 'VALID':
                self.pad = 0
                self.out_pad = 0
            else:
                self.pad = (self.size + (self.dilation_rate-1) * ( self.size-1 ))//2 - (1 - self.size%2)
                self.out_pad = self.stride - 1         
            self.kernel_size = (inchannel, self.outchn // self.groups, self.size, self.size)
        else:
            raise NotImplementedError("Deconv kernel only supports int")

    def forward(self, x: Tensor) -> Tensor:
        inh, inw = x.shape[2], x.shape[3]
        weight = self.weight
        if self._quant:
            x = self.input_quantizer(x)
            weight = self.w_quantizer(weight)
        x = F.conv_transpose2d(x, weight, self.bias, self.stride, self.pad, self.out_pad, self.groups, self.dilation_rate)
        outh, outw = x.shape[2], x.shape[3]
        if self.pad_mode=='SAME_LEFT':
            if outh!=inh*self.stride or outw!=inw*self.stride:
                x = x[:,:,:inh*self.stride,:inw*self.stride]
        return x 
    
    def to_torch(self) -> nn.ConvTranspose2d:
        conv = nn.ConvTranspose2d(in_channels = self.inchn, out_channels = self.outchn, kernel_size = self.kernel_size[2:], stride = self.stride,       # type: ignore
                        padding = self.pad, output_padding = self.out_pad, padding_mode = 'zeros', dilation = self.dilation_rate, groups = self.groups, bias = self.usebias)
        conv.weight.data[:] = self.weight.data[:]
        if (self.bias is not None) and (conv.bias is not None):
            conv.bias.data[:] = self.bias.data[:]
        return conv 


class conv1D(ConvBase):
    size: int
    kernel_size: Tuple[int,int,int]
    pad: int
    stride: int

    def __init__(self, size:int, outchn:int, stride:int=1, pad:PadModes='SAME_LEFT', dilation_rate:int=1, usebias:bool=True, groups:int=1):
        super().__init__(size=size, outchn=outchn, stride=stride, pad=pad, dilation_rate=dilation_rate, usebias=usebias, groups=groups)

    def initialize(self, size:int, outchn:int, stride:int=1, pad:PadModes='SAME_LEFT', dilation_rate:int=1, usebias:bool=True, groups:int=1) -> None:
        super().initialize(size=size, outchn=outchn, stride=stride, pad=pad, dilation_rate=dilation_rate, usebias=usebias, groups=groups)

    def _parse_args(self, input_shape):
        inchannel = input_shape[1]
        # parse args
        if self.pad_mode == 'VALID':
            self.pad = 0
        else:
            self.pad = (self.size + (self.dilation_rate-1) * ( self.size-1 ))//2
        self.kernel_size = (self.outchn, inchannel // self.gropus, self.size)

    def forward(self, x: Tensor) -> Tensor:
        return F.conv1d(x, self.weight, self.bias, self.stride, self.pad, self.dilation_rate, self.groups)
    
    def to_torch(self) -> nn.Conv1d:
        conv = nn.Conv1d(in_channels = self.inchn, out_channels = self.outchn, kernel_size = self.kernel_size[2:], stride = self.stride,        # type: ignore
                        padding = self.pad, padding_mode = 'zeros', dilation = self.dilation_rate, groups = self.groups, bias = self.usebias)
        conv.weight.data[:] = self.weight.data[:]
        if (self.bias is not None) and (conv.bias is not None):
            conv.bias.data[:] = self.bias.data[:]
        return conv 


class conv3D(ConvBase):
    size: TypeKSize3D
    stride: TypeKSize3D
    kernel_size: Tuple[int,int,int,int,int]
    pad: Union[int, Tuple[int,int,int]]

    def __init__(self, size:TypeKSize3D, outchn:int, stride:TypeKSize3D=1, pad:PadModes='SAME_LEFT', dilation_rate:int=1, usebias:bool=True, groups:int=1):
        super().__init__(size=size, outchn=outchn, stride=stride, pad=pad, dilation_rate=dilation_rate, usebias=usebias, groups=groups)

    def initialize(self, size:TypeKSize3D, outchn:int, stride:TypeKSize3D=1, pad:PadModes='SAME_LEFT', dilation_rate:int=1, usebias:bool=True, groups:int=1) -> None:
        super().initialize(size=size, outchn=outchn, stride=stride, pad=pad, dilation_rate=dilation_rate, usebias=usebias, groups=groups)
    
    def _parse_args(self, input_shape):
        inchannel = input_shape[1]
        # parse args
        if isinstance(self.size,list) or isinstance(self.size, tuple):
            if self.pad == 'VALID':
                self.pad = 0
            else:
                self.pad = ((self.size[0]+ (self.dilation_rate-1) * ( self.size[0]-1 ))//2, (self.size[1]+ (self.dilation_rate-1) * ( self.size[1]-1 ))//2, (self.size[2]+ (self.dilation_rate-1) * ( self.size[2]-1 ))//2)
            self.size = [self.outchn, inchannel // self.gropus, self.size[0], self.size[1], self.size[2]]
        else:
            if self.pad == 'VALID':
                self.pad = 0
            else:
                self.pad = (self.size + (self.dilation_rate-1) * ( self.size-1 ))//2
            self.kernel_size = (self.outchn, inchannel // self.gropus, self.size, self.size, self.size)

    def forward(self, x: Tensor) -> Tensor:
        return F.conv3d(x, self.weight, self.bias, self.stride, self.pad, self.dilation_rate, self.groups)
    
    def to_torch(self) -> nn.Conv3d:
        conv = nn.Conv3d(in_channels = self.inchn, out_channels = self.outchn, kernel_size = self.kernel_size[2:], stride = self.stride,       # type: ignore
                        padding = self.pad, padding_mode = 'zeros', dilation = self.dilation_rate, groups = self.groups, bias = self.usebias)
        conv.weight.data[:] = self.weight.data[:]
        if (self.bias is not None) and (conv.bias is not None):
            conv.bias.data[:] = self.bias.data[:]
        return conv 



from dataclasses import dataclass
from typing import Type, Union, Dict

from ..Consts.Types import *


##### START: QuantType classes 
@dataclass 
class QTypeBase():
    max_val: int 
    min_val: int 
    signed: bool


class QUint8(QTypeBase):
    max_val = 2 ** 8 -1 
    min_val = 0 
    signed = False


class QInt8(QTypeBase):
    max_val = 2 ** 7 - 1 
    min_val = - 2 ** 7 
    signed = True 


class QInt16(QTypeBase):
    max_val = 2 ** 15 - 1 
    min_val = - 2 ** 15 
    signed = True 


QTYPES: Dict[Union[QBitTypes, str], Type[QTypeBase]] = {"uint8": QUint8, "int8": QInt8, 'int16':QInt16}
##### END: QuantType classes 




import math
from typing import Union

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.nn.init as init
from torch import Tensor
from torch.nn.parameter import Parameter

from ..Base import Model
from ..Consts.Types import *
from ..Quant import QQuantizers, QuantizerBase


class fcLayer(Model):
    outsize: int 
    usebias: bool
    norm: bool
    insize: int
    weight: Parameter
    bias: Union[Parameter, None]
    input_quantizer: QuantizerBase
    w_quantizer: QuantizerBase

    def __init__(self, outsize: int, usebias:bool=True, norm:bool=False):
        super().__init__(outsize=outsize, usebias=usebias, norm=norm)

    def initialize(self, outsize: int, usebias:bool=True, norm:bool=False):
        self.outsize = outsize
        self.usebias = usebias
        self.norm = norm
        # self.bias = None 

    def build(self, *inputs):
        self.insize = inputs[0].shape[-1]
        self.weight = Parameter(torch.Tensor(self.outsize, self.insize))
        if self.usebias:
            self.bias = Parameter(torch.Tensor(self.outsize))
        else:
            self.register_parameter('bias', None)
        self.reset_params()

        if self._quant:
            bit_type: QBitTypes = self.get_flag('QActBit')
            if bit_type is None:
                bit_type = 'int8'
            obs_type: QObserverTypes = self.get_flag('QActObserver')
            if obs_type is None:
                obs_type = 'minmax'
            self.input_quantizer = QQuantizers['uniform'](zero_offset=False, bit_type=bit_type, observer=obs_type)
            self.w_quantizer = QQuantizers['uniform'](zero_offset=True, mode='channel_wise', is_weight=True)

    def set_input_quantizer(self, quantizer: QuantizerBase):
        if quantizer is not None:
            assert isinstance(quantizer, QuantizerBase), 'Must assign quantizer'
            self.input_quantizer = quantizer

    def reset_params(self):
        if self.get_flag('fc_init_mode')=='normal':
            init.normal_(self.weight, std=0.001)
            if self.bias is not None:
                init.zeros_(self.bias)
        else:
            init.kaiming_uniform_(self.weight, a=math.sqrt(5))
            if self.bias is not None:
                fan_in, _ = init._calculate_fan_in_and_fan_out(self.weight)
                bound = 1 / math.sqrt(fan_in) if fan_in > 0 else 0
                init.uniform_(self.bias, -bound, bound)

    def forward(self, x: Tensor) -> Tensor:
        if self.norm:
            with torch.no_grad():
                norm = x.norm(p=2, dim=1, keepdim=True)
                wnorm = self.weight.norm(p=2,dim=1, keepdim=True)
            x = x / norm
            weight = self.weight / wnorm
        else:
            weight = self.weight

        if self._quant:
            x = self.input_quantizer(x).contiguous()
            weight = self.w_quantizer(weight).contiguous()
        return F.linear(x, weight, self.bias)
    
    def __call__(self, x:Tensor) -> Tensor:
        return super().__call__(x)

    def to_torch(self) -> nn.Linear:
        fc = nn.Linear(self.insize, self.outsize, self.usebias)
        fc.weight.data[:] = self.weight.data[:]
        if self.bias is not None:
            fc.bias.data[:] = self.bias.data[:]
        return fc 



import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.nn.init as init
from torch import Tensor
from torch.nn.parameter import Parameter
from typing import Tuple 

from ..Base import Model


class NNUpSample(Model):
    inchannel: int 
    scale: int 
    size: Tuple[int,int,int,int]
    
    def __init__(self, scale:int):
        super().__init__(scale=scale)

    def initialize(self, scale: int):
        self.scale = scale

    def _parse_args(self, input_shape):
        self.inchannel = input_shape[1]
        self.size = (self.inchannel, 1, self.scale, self.scale)

    def build(self, *inputs):
        inp = inputs[0]
        self._parse_args(inp.shape)
        self.weight = Parameter(torch.Tensor(*self.size), requires_grad=False)
        self.reset_params()

    def reset_params(self):
        init.ones_(self.weight)

    def forward(self, x):
        return F.conv_transpose2d(x, self.weight, None, self.scale, 0, 0, self.inchannel, 1)
    
    def __call__(self, x:Tensor) -> Tensor:
        return super().__call__(x)
    
    def to_torch(self) -> nn.ConvTranspose2d:
        conv = nn.ConvTranspose2d(self.inchannel, self.inchannel, self.scale, self.scale, 0, 0, self.inchannel, bias=False)
        conv.weight.data[:] = self.weight.data[:]
        return conv 
    
    def _load_from_state_dict(self, state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys, error_msgs):
        pass

    def _save_to_state_dict(self, destination, prefix, keep_vars):
        pass


class PadAround2D(nn.Module):
    def __init__(self, n_pix: int):
        super().__init__()
        self.n_pix = n_pix

    def forward(self, x: Tensor) -> Tensor: 
        return F.pad(x, (self.n_pix, self.n_pix, self.n_pix, self.n_pix))
    
    def __call__(self, x: Tensor) -> Tensor:
        return super().__call__(x)


class BilinearUpSample(Model):
    factor: int 
    pad0: int 
    inchannel: int 
    weight: Parameter

    def __init__(self, factor: int):
        super().__init__(factor)

    def initialize(self, factor: int):
        self.factor = factor
        self.pad0 = factor//2 * 3 + factor%2

    def build(self, *inputs):
        inp = inputs[0]
        self.inchannel = inp.shape[1]
        filter_size = 2*self.factor - self.factor%2
        k = self.upsample_kernel(filter_size)
        k = k[None,...]
        k = np.repeat(k, self.inchannel, axis=0)
        self.weight = Parameter(torch.from_numpy(k), requires_grad=False)

    def upsample_kernel(self,size):
        factor = (size +1)//2
        if size%2==1:
            center = factor - 1
        else:
            center = factor - 0.5
        og = np.ogrid[:size, :size]
        k = (1 - abs(og[0]-center)/factor) * (1-abs(og[1]-center)/factor)
        return np.array(k, dtype=np.float32)

    def forward(self, x):
        x = F.pad(x, (1,1,1,1), 'replicate')
        x = F.conv_transpose2d(x, self.weight, None, self.factor, self.pad0, 0, self.inchannel, 1)
        return x 
    
    def __call__(self, x: Tensor) -> Tensor:
        return super().__call__(x)
    
    def to_torch(self) -> nn.Sequential:
        conv = nn.ConvTranspose2d(self.inchannel, self.inchannel, self.factor, self.factor, 0, 0, self.inchannel, bias=False)
        conv.weight.data[:] = self.weight.data[:]
        mod = nn.Sequential(PadAround2D(1), conv)
        return mod 

    def _load_from_state_dict(self, state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys, error_msgs):
        pass

    def _save_to_state_dict(self, destination, prefix, keep_vars):
        pass



from typing import Any, Union

import yaml


class ConfigDict():
	'''
	Easy-to-use dict
	Python dictionary: d[A][B]  ->  ConfigDict:  d.A.B
	'''
	def __init__(self, d: dict):
		self.dict_str = str(d)
		self.dict = self.parse_dict(d)

	@staticmethod
	def parse_dict(d):
		if isinstance(d, dict):
			res = {}
			for k in d:
				if isinstance(d[k], dict):
					buff = ConfigDict(d[k])
				else:
					buff = ConfigDict.parse_dict(d[k])
				res[k] = buff
		elif isinstance(d, list):
			res = [ConfigDict.parse_dict(i) for i in d]
		else:
			res = d
		return res 

	def __getattr__(self, key) -> Any:
		if not key[0]=='_':
			return self.dict[key]
		else:
			super().__getattr__(key)             # type: ignore

	def __str__(self) -> str:
		return self.dict_str


def load_yaml(f: str) -> ConfigDict:
	with open(f) as file:
		config: dict = yaml.safe_load(file)
	return ConfigDict(config) 


if __name__=='__main__':
	# for debug purpose
	d = {'AB': 1, 'BC':[2,3], 'CD': {'AAA':1, 'BBB':2}}
	dd = ConfigDict(d)
	print(dd.CD.AAA)

	d2 = load_yaml('../abc.yaml')
	print(d2.MODEL.AAB)



from .Observers import (MinMaxObserver, ObserverBase, OmseObserver,
                        PercentileObserver, QObservers)
from .QAct import QAct
from .QATop import QATFunc
from .QTypes import QTYPES, QInt8, QInt16, QTypeBase, QUint8
from .Quantizers import QQuantizers, QuantizerBase, UniformQuantizer
from .QCalibrator import LayerCalibrator
from .QLayers import QMatmul, QAdd



from typing import Union, Tuple

import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor

from ..Base import Model
from ..Consts.Types import *


## Pool layers 
class MaxPool2d(Model):
    size: TypeKSize2D
    stride: TypeKSize2D
    pad_mode: PadModes
    dilation_rate: int
    pad: Union[int, Tuple[int,int]]

    def __init__(self, size:TypeKSize2D, stride:TypeKSize2D=1, pad:PadModes='SAME_LEFT', dilation_rate:int=1):
        super().__init__(size, stride, pad, dilation_rate)

    def initialize(self, size:TypeKSize2D, stride:TypeKSize2D=1, pad:PadModes='SAME_LEFT', dilation_rate:int=1):
        self.size = size
        self.stride = stride
        self.pad_mode = pad
        self.dilation_rate = dilation_rate
        
    def build(self, x:Tensor):
        self._parse_args()

    def _parse_args(self):
        # parse args
        if isinstance(self.size,list) or isinstance(self.size, tuple):
            if self.pad_mode == 'VALID':
                self.pad = 0
            else:
                self.pad = (self.size[0]//2, self.size[1]//2)
        else:
            if self.pad_mode == 'VALID':
                self.pad = 0
            else:
                self.pad = self.size//2

    def forward(self, x: Tensor) -> Tensor:
        return F.max_pool2d(x, self.size, self.stride, self.pad, self.dilation_rate, False, False)
    
    def __call__(self, x: Tensor) -> Tensor:
        return super().__call__(x)
    
    def to_torch(self) -> nn.MaxPool2d:
        pool = nn.MaxPool2d(self.size, self.stride, padding=self.pad, dilation=self.dilation_rate) #type: ignore
        return pool


class AvgPool2d(Model):
    size: TypeKSize2D
    stride: TypeKSize2D
    pad_mode: PadModes
    dilation_rate: int
    pad: Union[int, Tuple[int,int]]

    def __init__(self, size:TypeKSize2D, stride:TypeKSize2D=1, pad:PadModes='SAME_LEFT', dilation_rate:int=1):
        super().__init__(size, stride, pad, dilation_rate)

    def initialize(self, size:TypeKSize2D, stride:TypeKSize2D=1, pad:PadModes='SAME_LEFT', dilation_rate:int=1):
        self.size = size
        self.stride = stride
        self.pad_mode = pad
        self.dilation_rate = dilation_rate
        
    def build(self, x:Tensor):
        self._parse_args()

    def _parse_args(self):
        # parse args
        if isinstance(self.size,list) or isinstance(self.size, tuple):
            if self.pad_mode == 'VALID':
                self.pad = 0
            else:
                self.pad = (self.size[0]//2, self.size[1]//2)
        else:
            if self.pad_mode == 'VALID':
                self.pad = 0
            else:
                self.pad = self.size//2

    def forward(self, x: Tensor) -> Tensor:
        return F.avg_pool2d(x, self.size, self.stride, self.pad, self.dilation_rate, False, False)
    
    def __call__(self, x: Tensor) -> Tensor:
        return super().__call__(x)
    
    def to_torch(self) -> nn.AvgPool2d:
        pool = nn.AvgPool2d(self.size, self.stride, padding=self.pad, dilation=self.dilation_rate) #type: ignore
        return pool



from .Utils import Config
from . import Base, Model, Quant, Tools

sul_tool = Tools

# modify loguru format. Add \n in front, in case other modules overrides stderr
import sys 
from loguru import logger 


class PrintStream():
    def write(self, message):
        print(message, end='')


logger.remove()
fmt = '<g>{time:YYYY-MM-DD HH:mm:ss}</g> | PID: <c>{process}</c> | <lvl>{level}</lvl> - <lvl>{message}</lvl> '
stream_obj = PrintStream()
# logger.add(sys.stderr, format=fmt)
logger.add(stream_obj, format=fmt, colorize=True)





from __future__ import annotations

import multiprocessing as mp
from abc import ABC, abstractmethod
from typing import Any, Type

from .Progress import progress_bar


class VisualizerBase(mp.Process, ABC):
    def __init__(self, imgListQ: mp.Queue, progressQ: mp.Queue):
        super().__init__()
        self.imgListQ = imgListQ
        self.progressQ = progressQ
        self.initialize()
    
    def initialize(self):
        ...
    
    def run(self):
        while self.imgListQ.qsize() > 0:
            data_bundle = self.imgListQ.get()
            self.process_img(*data_bundle)
            self.progressQ.put(-1)

    @abstractmethod
    def process_img(self, *data_bundle):
        ...


class ProgressProc(mp.Process):
    def __init__(self, progressQ: mp.Queue):
        super().__init__()
        self.prog = progress_bar()
        self.progressQ = progressQ

    def run(self):
        self.prog.start()
        total = self.progressQ.get()
        task = self.prog.add_task('Visualizing', total=total)
        cnt = 0
        while 1:
            self.progressQ.get()
            self.prog.advance(task)
            cnt += 1 
            if cnt == total:
                break 
        self.prog.stop()


def start_visualize(data_bundle: list[Any], TVisProc: Type[VisualizerBase], n_procs: int=4):
    progressQ = mp.Queue()
    imageListQ = mp.Queue()

    for d in data_bundle:
        imageListQ.put(d)

    progressQ.put(len(data_bundle))

    progress_proc = ProgressProc(progressQ)
    progress_proc.start()

    procs = []
    for _ in range(n_procs):
        proc = TVisProc(imageListQ, progressQ)
        proc.start()
        procs.append(proc)
    
    for p in procs:
        p.join()




import torch
import torch.nn.functional as F
from torch import Tensor

from ..Base import Model


def activation(x: Tensor, act, **kwargs) -> Tensor:
    if act==-1:
        return x
    elif act==0:
        return F.relu(x)
    elif act==1:
        return F.leaky_relu(x, negative_slope=0.1)
    elif act==2:
        return F.elu(x)
    elif act==3:
        return F.tanh(x)
    elif act==6:
        return torch.sigmoid(x)
    elif act==10:
        return F.gelu(x)
    elif act==11:
        return F.silu(x)
    else:
        raise NotImplementedError(f'Activation [{act}] is not supported')


class Activation(Model):
    act_num: int

    def __init__(self, act: int):
        super().__init__(act)

    def initialize(self, act: int):
        self.act_num = act 
        if act==9:
            self.act = torch.nn.PReLU(num_parameters=1)
    
    def build(self, x:Tensor):
        if self.act_num==8:
            self.act = torch.nn.PReLU(num_parameters=x.shape[-1]) 

    def forward(self, x: Tensor) -> Tensor:
        if self.act_num==8 or self.act_num==9:
            return self.act(x)
        else:
            return activation(x, self.act_num)

    def __call__(self, x: Tensor) -> Tensor:
        return super().__call__(x)




import glob
import os
import re
import shutil
from typing import Iterator, Union

from loguru import logger


# path utils 
class Path(os.PathLike[str]):
    def __init__(self, path: str):
        self.path = path 

    def __str__(self) -> str:
        return self.path
    
    def __fspath__(self) -> str:
        return self.path 

    @staticmethod
    def _tostr(obj: Union[str, 'Path']) -> str:
        assert isinstance(obj, (str, Path))
        if isinstance(obj, str):
            return obj 
        else:
            return obj.path 

    def __add__(self, other: Union[str, 'Path']) -> 'Path':
        path_new = os.path.join(self.path, self._tostr(other))
        return Path(path_new)
    
    def __radd__(self, other: Union[str, 'Path']) -> 'Path':
        path_new = os.path.join(self._tostr(other), self.path)
        return Path(path_new)
    
    def __contains__(self, s: str):
        return s in self.path
    
    def __getitem__(self, idx: int) -> str:
        return self.path.split('/')[idx]

    def find_one(self, pattern: str, use_regex: bool = False) -> 'Path':
        if not use_regex:
            pattern = '^' + pattern.replace('*', '.*') + '$'
        new_path = None 
        for item in glob.glob(os.path.join(self.path, '*')):
            result = re.search(pattern, item.split('/')[-1])
            if result is not None:
                new_path = item 
        if new_path is not None:
            return Path(new_path)
        else:
            raise FileNotFoundError(f'Cannot find pattern {pattern} in folder {self.path}')
        
    def find(self, pattern: str, use_regex: bool = False) -> Iterator['Path']:
        if not use_regex:
            pattern = '^' + pattern.replace('*', '.*') + '$'
        for item in glob.glob(os.path.join(self.path, '*')):
            result = re.search(pattern, item.split('/')[-1])
            if result is not None:
                yield Path(item)

    def auto(self) -> 'Path':
        sub_paths = glob.glob(os.path.join(self.path, '*'))
        if len(sub_paths)==0:
            raise FileNotFoundError(f'Auto: No file under {self.path}')
        else:
            return Path(sub_paths[0])
        
    def copy_to(self, target: 'Path'):
        os.makedirs(os.path.dirname(target.path), exist_ok=True)
        shutil.copy(self.path, target.path)
    
    def copy_dir_to(self, target: 'Path'):
        os.makedirs(os.path.dirname(target.path), exist_ok=True)
        if os.path.exists(target.path):
            target.remove_dir()
        shutil.copytree(self.path, target.path)

    def remove(self):
        try:
            os.remove(self.path)
        except Exception as e:
            logger.warning(f'Cannot remove {self.path}. {e}')
    
    def remove_dir(self):
        try:
            shutil.rmtree(self.path)
        except Exception as e:
            logger.warning(f'Cannot remove {self.path}. {e}')

    def makedirs(self):
        os.makedirs(self.path, exist_ok=True)
    
    def replace(self, src: str, tgt: str) -> 'Path':
        path_new = self.path.replace(src, tgt)
        return Path(path_new)

    def exists(self) -> bool:
        return os.path.exists(self.path)
    
    def basename(self) -> str:
        return os.path.basename(self.path)
    
    def dirname(self) -> 'Path':
        return Path(os.path.dirname(self.path))




import torch 
import torchvision.ops as ops

from typing import Optional
from typing_extensions import Literal, Self
from torch import Tensor 


# bbox utils 
class BBox():
    bbox: Tensor 
    conf: Optional[Tensor]
    box_format: Literal['xyxy', 'x1y1wh', 'xcycwh']

    def __init__(self, bbox: Tensor, box_format: Literal['xyxy', 'x1y1wh', 'xcycwh'], conf: Optional[Tensor]=None): 
        # dont use inplace
        assert (bbox.shape[-1] == 4) or (bbox.shape[-1] == 5)
        assert len(bbox.shape)>=1, 'BBox must has n_dim >= 1'
        self.bbox = bbox[..., :4].clone()
        if conf is None:
            if bbox.shape[-1] == 4:
                self.conf = None 
            else:
                self.conf = bbox[..., 4].clone()
        else:
            self.conf = conf.clone()
        self.box_format = box_format

    def copy(self):
        return type(self)(self.bbox, self.box_format, self.conf)
    
    @property
    def shape(self):
        return self.bbox.shape

    def x1y1wh2xyxy(self) -> Self:
        x2 = self.bbox[..., 0] + self.bbox[..., 2]
        y2 = self.bbox[..., 1] + self.bbox[..., 3]
        new_bbox = torch.stack([self.bbox[..., 0], self.bbox[..., 1], x2, y2], dim=-1)
        return type(self)(new_bbox, 'xyxy', self.conf)
    
    def xyxy2x1y1wh(self) -> Self:
        w = self.bbox[..., 2] - self.bbox[..., 0]
        h = self.bbox[..., 3] - self.bbox[..., 1]
        new_bbox = torch.stack([self.bbox[..., 0], self.bbox[..., 1], w, h], dim=-1)
        return type(self)(new_bbox, 'x1y1wh', self.conf)
    
    def xyxy2xcycwh(self) -> Self:
        xc = 0.5 * (self.bbox[..., 0] + self.bbox[..., 2])
        yc = 0.5 * (self.bbox[..., 1] + self.bbox[..., 3])
        w = self.bbox[..., 2] - self.bbox[..., 0]
        h = self.bbox[..., 3] - self.bbox[..., 1]
        new_bbox = torch.stack([xc,yc,w,h], dim=-1)
        return type(self)(new_bbox, 'xcycwh', self.conf)
    
    def xcycwh2xyxy(self) -> Self:
        x1 = self.bbox[..., 0] - 0.5 * self.bbox[..., 2]
        y1 = self.bbox[..., 1] - 0.5 * self.bbox[..., 3]
        x2 = self.bbox[..., 0] + 0.5 * self.bbox[..., 2]
        y2 = self.bbox[..., 1] + 0.5 * self.bbox[..., 3]
        new_bbox = torch.stack([x1, y1, x2, y2], dim=-1)
        return type(self)(new_bbox, 'xyxy', self.conf)
    
    def x1y1wh2xcycwh(self) -> Self:
        xc = self.bbox[..., 0] + 0.5 * self.bbox[..., 2]
        yc = self.bbox[..., 1] + 0.5 * self.bbox[..., 3]
        new_bbox = torch.stack([xc, yc, self.bbox[..., 2], self.bbox[..., 3]], dim=-1)
        return type(self)(new_bbox, 'xcycwh', self.conf)
    
    def xcycwh2x1y1wh(self) -> Self:
        x1 = self.bbox[..., 0] - 0.5 * self.bbox[..., 2]
        y1 = self.bbox[..., 1] - 0.5 * self.bbox[..., 3] 
        new_bbox = torch.stack([x1, y1, self.bbox[..., 2], self.bbox[..., 3]], dim=-1)
        return type(self)(new_bbox, 'x1y1wh', self.conf)
    
    def convert(self, target_format: Literal['xyxy', 'x1y1wh', 'xcycwh']) -> Self:
        funcs = {('x1y1wh', 'xcycwh'): self.x1y1wh2xcycwh, 
                 ('x1y1wh', 'xyxy'): self.x1y1wh2xyxy,
                 ('xcycwh', 'x1y1wh'): self.xcycwh2x1y1wh,
                 ('xcycwh', 'xyxy'): self.xcycwh2xyxy,
                 ('xyxy', 'x1y1wh'): self.xyxy2x1y1wh,
                 ('xyxy', 'xcycwh'): self.xyxy2xcycwh}
        
        convert_tuple = (self.box_format, target_format)
        if convert_tuple in funcs:
            new_obj = funcs[convert_tuple]()
        else:
            new_obj = self.copy()
        return new_obj

    def iou(self, other: Self) -> Tensor:
        b1 = self.convert('xyxy')
        b2 = other.convert('xyxy')
        return ops.box_iou(b1.bbox, b2.bbox)

    def __xor__(self, other: Self) -> Tensor:
        return self.iou(other)

    def distance(self, other: Self) -> Tensor:
        b1 = self.convert('xcycwh')
        b2 = other.convert('xcycwh')
        assert len(self.shape) in [2,3], 'Box shape must be 2 or 3 to compute distance'
        if len(self.shape)==3:
            assert self.shape[0]==other.shape[0], f'Dim 0 should has the same shape for bbox distance. Got {self.shape[0]} and {other.shape[0]}'
        return torch.cdist(b1.bbox[..., :2], b2.bbox[..., :2])

    def __or__(self, other: Self) -> Tensor:
        return self.distance(other)

    def cat(self, other: Self, dim: int = 1) -> Self: 
        assert self.box_format==other.box_format, 'box format should be the same for concatenation'
        new_box = torch.cat([self.bbox, other.bbox], dim=0)
        if (self.conf is not None) and (other.conf is not None):
            new_conf = torch.cat([self.conf, other.conf], dim=0)
        else:
            new_conf = None 
        return BBOX(new_box, self.box_format, new_conf)

    def __matmul__(self, other: Self) -> Self:
        return self.cat(other)

    def __getitem__(self, idx: Tensor) -> Self:
        new_box = self.bbox[idx]
        if self.conf is not None:
            new_conf = self.conf[idx]
        else:
            new_conf = None 
        return BBOX(new_box, self.box_format, new_conf)

    def __len__(self):
        return len(self.bbox)

    def inside(self, other: Self) -> Tensor:
        assert len(self)==len(other), f'Length should be the same for bbox __contain__, first: {len(other)}, second: {len(self)}'  # TODO: this assertion should be updated 
        this_box = self.convert('xcycwh').bbox
        other_box = other.convert('xyxy').bbox
        xs = this_box[..., 0]
        ys = this_box[..., 1]
        result = ((xs > other_box[..., 0]) & (xs < other_box[..., 2]) & (ys > other_box[..., 1]) & (ys < other_box[..., 3]))
        return result



from .BBox import BBox
from .Path import Path
from .Progress import progress_bar
from .Video import (VideoSaver, check_fps, check_frame_num, combine_audio,
                    compress_video, extract_frames)
from .Visualizer import VisualizerBase, start_visualize
